///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2012 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Threading;

using com.espertech.esper.client;
using com.espertech.esper.client.scopetest;
using com.espertech.esper.client.time;
using com.espertech.esper.compat.collections;
using com.espertech.esper.compat.logging;
using com.espertech.esper.support.bean;
using com.espertech.esper.support.client;

using NUnit.Framework;

namespace com.espertech.esper.regression.context
{
    [TestFixture]
    public class TestContextMTSegmented 
    {
        private static readonly ILog Log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
    
        private EPServiceProvider _epService;
        private SupportUpdateListener _listener;
    
        [SetUp]
        public void SetUp()
        {
            Configuration config = SupportConfigFactory.GetConfiguration();
            config.AddEventType("SupportBean", typeof(SupportBean));
            _epService = EPServiceProviderManager.GetDefaultProvider(config);
            _epService.Initialize();
            _listener = new SupportUpdateListener();
        }
    
        [TearDown]
        public void TearDown() {
            _listener = null;
        }
    
        [Test]
        public void TestSegmentedContext()
        {
            String[] choices = "A,B,C,D".Split(',');
            TrySend(4, 1000, choices);
        }
    
        private void TrySend(int numThreads, int numEvents, String[] choices)
        {
            if (numEvents < choices.Length) {
                throw new ArgumentException("Number of events must at least match number of choices");
            }
            
            _epService.EPRuntime.SendEvent(new CurrentTimeEvent(0));
            _epService.EPAdministrator.CreateEPL("create variable boolean myvar = false");
            _epService.EPAdministrator.CreateEPL("create context SegmentedByString as partition by TheString from SupportBean");
            EPStatement stmt = _epService.EPAdministrator.CreateEPL("context SegmentedByString select TheString, count(*) - 1 as cnt from SupportBean output snapshot when myvar = true");
            stmt.Events += _listener.Update;
    
            // preload - since concurrently sending same-category events an event can be dropped
            for (int i = 0; i < choices.Length; i++) {
                _epService.EPRuntime.SendEvent(new SupportBean(choices[i], 0));
            }
    
            EventRunnable[] runnables = new EventRunnable[numThreads];
            for (int i = 0; i < runnables.Length; i++) {
                runnables[i] = new EventRunnable(_epService, numEvents, choices);
            }
    
            // start
            Thread[] threads = new Thread[runnables.Length];
            for (int i = 0; i < runnables.Length; i++) {
                threads[i] = new Thread(runnables[i].Run);
                threads[i].Start();
            }
    
            // join
            Log.Info("Waiting for completion");
            for (int i = 0; i < runnables.Length; i++) {
                threads[i].Join();
            }
    
            IDictionary<string, long> totals = new Dictionary<string, long>();
            foreach (var choice in choices) {
                totals.Put(choice, 0L);
            }
    
            // verify
            int sum = 0;
            for (int i = 0; i < runnables.Length; i++) {
                Assert.IsNull(runnables[i].Exception);
                foreach (var entry in runnables[i].Totals) {
                    var current = totals.Get(entry.Key);
                    current += entry.Value;
                    sum += entry.Value;
                    totals.Put(entry.Key, current);
                    //Console.Out.WriteLine("Thread " + i + " key " + entry.Key + " count " + entry.Value);
                }
            }
    
            Assert.AreEqual(numThreads * numEvents, sum);
    
            _epService.EPRuntime.SetVariableValue("myvar", true);
            _epService.EPRuntime.SendEvent(new CurrentTimeEvent(10000));
            EventBean[] result = _listener.LastNewData;
            Assert.AreEqual(choices.Length, result.Length);
            foreach (EventBean item in result) {
                var stringValue = (String) item.Get("TheString");
                var count = (long?) item.Get("cnt");
                //Console.Out.WriteLine("String " + string + " count " + count);
                Assert.AreEqual(count, totals.Get(stringValue));
            }
        }
    
        public class EventRunnable 
        {
            private readonly EPServiceProvider _epService;
            private readonly int _numEvents;
            private readonly String[] _choices;
            private readonly IDictionary<string, int> _totals = new Dictionary<string, int>();
    
            private Exception _exception;
    
            public EventRunnable(EPServiceProvider epService, int numEvents, String[] choices) 
            {
                _epService = epService;
                _numEvents = numEvents;
                _choices = choices;
            }
    
            public void Run() {
                Log.Info("Started event send");
    
                try {
                    for (int i = 0; i < _numEvents; i++) {
                        String chosen = _choices[i % _choices.Length];
                        _epService.EPRuntime.SendEvent(new SupportBean(chosen, 1));

                        var current = _totals.Get(chosen, 0);
                        current += 1;
                        _totals.Put(chosen, current);
                    }
                }
                catch (Exception ex) {
                    Log.Error("Exception encountered: " + ex.Message, ex);
                    _exception = ex;
                }
    
                Log.Info("Completed event send");
            }

            public Exception Exception
            {
                get { return _exception; }
            }

            public IDictionary<string, int> Totals
            {
                get { return _totals; }
            }
        }
    }
}
