///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2012 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Reflection;
using System.Threading;

using com.espertech.esper.client;
using com.espertech.esper.client.time;
using com.espertech.esper.compat.logging;
using com.espertech.esper.support.bean;
using com.espertech.esper.support.client;

using NUnit.Framework;

namespace com.espertech.esper.regression.context
{
    /// <summary>
    /// Test for multithread-safety (or lack thereof) for iterators:
    ///      iterators fail with concurrent mods as expected behavior
    /// </summary>
    [TestFixture]
    public class TestContextMTTemporalStartStop
    {
        private static readonly ILog Log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

        private EPServiceProvider _epService;

        [SetUp]
        public void SetUp()
        {
            Configuration config = SupportConfigFactory.GetConfiguration();
            config.AddEventType("SupportBean", typeof(SupportBean));
            _epService = EPServiceProviderManager.GetDefaultProvider(config);
            _epService.Initialize();
        }

        [Test]
        public void TestMT()
        {
            TrySend();
        }

        private void TrySend()
        {
            _epService.EPAdministrator.CreateEPL(
                "create context EverySecond as start (*, *, *, *, *, *) end (*, *, *, *, *, *)");
            _epService.EPAdministrator.CreateEPL("context EverySecond select * from SupportBean");

            var timerRunnable = new TimerRunnable(_epService, 0, 24*60*60*1000, 1000);
            var timerThread = new Thread(timerRunnable.Run) {Name = "timer"};

            var eventRunnable = new EventRunnable(_epService, 1000000);
            var eventThread = new Thread(eventRunnable.Run) {Name = "event"};

            timerThread.Start();
            eventThread.Start();

            timerThread.Join();
            eventThread.Join();
            Assert.IsNull(eventRunnable.Exception);
            Assert.IsNull(timerRunnable.Exception);
        }

        public class TimerRunnable
        {
            private readonly long _end;
            private readonly EPServiceProvider _service;
            private readonly long _increment;
            private readonly long _start;

            private Exception _exception;

            public TimerRunnable(EPServiceProvider service,
                                 long start,
                                 long end,
                                 long increment)
            {
                _service = service;
                _start = start;
                _end = end;
                _increment = increment;
            }

            public Exception Exception
            {
                get { return _exception; }
            }

            public void Run()
            {
                Log.Info("Started time drive");
                try
                {
                    long current = _start;
                    long stepCount = 0;
                    long expectedSteps = (_end - _start)/_increment;
                    while (current < _end)
                    {
                        _service.EPRuntime.SendEvent(new CurrentTimeEvent(current));
                        current += _increment;
                        stepCount++;

                        if (stepCount%10000 == 0)
                        {
                            Log.Info("Sending step #" + stepCount + " of " + expectedSteps);
                        }
                    }
                }
                catch (Exception ex)
                {
                    Log.Error("Exception encountered: " + ex.Message, ex);
                    _exception = ex;
                }
                Log.Info("Completed time drive");
            }
        }

        public class EventRunnable
        {
            private readonly EPServiceProvider _service;
            private readonly long _numEvents;

            private Exception _exception;

            public EventRunnable(EPServiceProvider service,
                                 long numEvents)
            {
                _service = service;
                _numEvents = numEvents;
            }

            public Exception Exception
            {
                get { return _exception; }
            }

            public void Run()
            {
                Log.Info("Started event send");
                try
                {
                    long count = 0;
                    while (count < _numEvents)
                    {
                        _service.EPRuntime.SendEvent(new SupportBean());
                        count++;

                        if (count%10000 == 0)
                        {
                            Log.Info("Sending event #" + count);
                        }
                    }
                }
                catch (Exception ex)
                {
                    Log.Error("Exception encountered: " + ex.Message, ex);
                    _exception = ex;
                }
                Log.Info("Completed event send");
            }
        }
    }
}