///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2012 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;

using com.espertech.esper.client;
using com.espertech.esper.client.scopetest;
using com.espertech.esper.client.soda;
using com.espertech.esper.compat.collections;
using com.espertech.esper.support.bean;
using com.espertech.esper.support.client;
using com.espertech.esper.util;

using NUnit.Framework;

namespace com.espertech.esper.regression.view
{
    [TestFixture]
    public class TestCastExpr 
    {
        private EPServiceProvider _epService;
        private SupportUpdateListener _listener;
    
        [SetUp]
        public void SetUp()
        {
            _listener = new SupportUpdateListener();
            _epService = EPServiceProviderManager.GetDefaultProvider(SupportConfigFactory.GetConfiguration());
            _epService.Initialize();
        }
    
        [TearDown]
        public void TearDown()
        {
            _listener = null;
        }
    
        [Test]
        public void TestCastSimple()
        {
            String stmtText = "select cast(TheString as string) as t0, " +
                              " cast(IntBoxed, int) as t1, " +
                              " cast(FloatBoxed, " + typeof(float).FullName + ") as t2, " +
                              " cast(TheString, System.String) as t3, " +
                              " cast(IntPrimitive, " + typeof(int).FullName + ") as t4, " +
                              " cast(IntPrimitive, long) as t5, " +
                              " cast(IntPrimitive, System.Object) as t6, " +
                              " cast(FloatBoxed, long) as t7 " +
                              " from " + typeof(SupportBean).FullName;
    
            EPStatement selectTestCase = _epService.EPAdministrator.CreateEPL(stmtText);
            selectTestCase.Events += _listener.Update;
    
            Assert.AreEqual(typeof(string), selectTestCase.EventType.GetPropertyType("t0"));
            Assert.AreEqual(typeof(int?), selectTestCase.EventType.GetPropertyType("t1"));
            Assert.AreEqual(typeof(float?), selectTestCase.EventType.GetPropertyType("t2"));
            Assert.AreEqual(typeof(string), selectTestCase.EventType.GetPropertyType("t3"));
            Assert.AreEqual(typeof(int?), selectTestCase.EventType.GetPropertyType("t4"));
            Assert.AreEqual(typeof(long?), selectTestCase.EventType.GetPropertyType("t5"));
            Assert.AreEqual(typeof(object), selectTestCase.EventType.GetPropertyType("t6"));
            Assert.AreEqual(typeof(long?), selectTestCase.EventType.GetPropertyType("t7"));
    
            SupportBean bean = new SupportBean("abc", 100);
            bean.FloatBoxed = 9.5f;
            bean.IntBoxed = 3;
            _epService.EPRuntime.SendEvent(bean);
            EventBean theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {"abc", 3, 9.5f, "abc", 100, 100L, 100, 9l});
    
            bean = new SupportBean(null, 100);
            bean.FloatBoxed = null;
            bean.IntBoxed = null;
            _epService.EPRuntime.SendEvent(bean);
            theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {null, null, null, null, 100, 100L, 100, null});
            bean = new SupportBean(null, 100);
            bean.FloatBoxed = null;
            bean.IntBoxed = null;
            _epService.EPRuntime.SendEvent(bean);
            theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {null, null, null, null, 100, 100L, 100, null});
    
            // test cast with chained
            selectTestCase.Dispose();
            stmtText = "select cast(one as " + typeof(SupportBean).FullName + ").get_TheString() as t0" +
                              " from " + typeof(SupportBeanObject).FullName;
            selectTestCase = _epService.EPAdministrator.CreateEPL(stmtText);
            selectTestCase.Events += _listener.Update;
    
            _epService.EPRuntime.SendEvent(new SupportBeanObject(new SupportBean("E1", 1)));
            Assert.AreEqual("E1", _listener.AssertOneGetNewAndReset().Get("t0"));
        }
    
        [Test]
        public void TestCastAsParse()
        {
            String stmtText = "select cast(TheString, int) as t0 from " + typeof(SupportBean).FullName;
            EPStatement selectTestCase = _epService.EPAdministrator.CreateEPL(stmtText);
            selectTestCase.Events += _listener.Update;
    
            Assert.AreEqual(typeof(int?), selectTestCase.EventType.GetPropertyType("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBean("12", 1));
            EPAssertionUtil.AssertProps(_listener.AssertOneGetNewAndReset(), "t0".Split(','), new Object[] {12});
        }
    
        [Test]
        public void TestCastDoubleAndNull_OM()
        {
            String stmtText = "select cast(item?,double) as t0 " +
                              "from " + typeof(SupportMarkerInterface).FullName;
    
            EPStatementObjectModel model = new EPStatementObjectModel();
            model.SelectClause = SelectClause.Create().Add(Expressions.Cast("item?", "double"), "t0");
            model.FromClause = FromClause.Create(FilterStream.Create(typeof(SupportMarkerInterface).FullName));
            model = (EPStatementObjectModel) SerializableObjectCopier.Copy(model);
            Assert.AreEqual(stmtText, model.ToEPL());
    
            EPStatement selectTestCase = _epService.EPAdministrator.Create(model);
            selectTestCase.Events += _listener.Update;
    
            Assert.AreEqual(typeof(double?), selectTestCase.EventType.GetPropertyType("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(100));
            Assert.AreEqual(100d, _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot((byte)2));
            Assert.AreEqual(2d, _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(77.7777));
            Assert.AreEqual(77.7777d, _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(6L));
            Assert.AreEqual(6d, _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(null));
            Assert.AreEqual(null, _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot("abc"));
            Assert.AreEqual(null, _listener.AssertOneGetNewAndReset().Get("t0"));
        }
    
        [Test]
        public void TestCastStringAndNull_Compile()
        {
            String stmtText = "select cast(item?,System.String) as t0 " +
                              "from " + typeof(SupportMarkerInterface).FullName;
    
            EPStatementObjectModel model = _epService.EPAdministrator.CompileEPL(stmtText);
            model = (EPStatementObjectModel) SerializableObjectCopier.Copy(model);
            EPStatement selectTestCase = _epService.EPAdministrator.Create(model);
            Assert.AreEqual(stmtText, model.ToEPL());
            selectTestCase.Events += _listener.Update;
    
            Assert.AreEqual(typeof(string), selectTestCase.EventType.GetPropertyType("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(100));
            Assert.AreEqual("100", _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot((byte)2));
            Assert.AreEqual("2", _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(77.7777));
            Assert.AreEqual("77.7777", _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(6L));
            Assert.AreEqual("6", _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(null));
            Assert.AreEqual(null, _listener.AssertOneGetNewAndReset().Get("t0"));
    
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot("abc"));
            Assert.AreEqual("abc", _listener.AssertOneGetNewAndReset().Get("t0"));
        }
    
        [Test]
        public void TestCastInterface()
        {
            String caseExpr = "select cast(item?, " + typeof(SupportMarkerInterface).FullName + ") as t0, " +
                              " cast(item?, " + typeof(ISupportA).FullName + ") as t1, " +
                              " cast(item?, " + typeof(ISupportBaseAB).FullName + ") as t2, " +
                              " cast(item?, " + typeof(ISupportBaseABImpl).FullName + ") as t3, " +
                              " cast(item?, " + typeof(ISupportC).FullName + ") as t4, " +
                              " cast(item?, " + typeof(ISupportD).FullName + ") as t5, " +
                              " cast(item?, " + typeof(ISupportAImplSuperG).FullName + ") as t6, " +
                              " cast(item?, " + typeof(ISupportAImplSuperGImplPlus).FullName + ") as t7 " +
                              " from " + typeof(SupportMarkerInterface).FullName;
    
            EPStatement selectTestCase = _epService.EPAdministrator.CreateEPL(caseExpr);
            selectTestCase.Events += _listener.Update;
    
            Assert.AreEqual(typeof(SupportMarkerInterface), selectTestCase.EventType.GetPropertyType("t0"));
            Assert.AreEqual(typeof(ISupportA), selectTestCase.EventType.GetPropertyType("t1"));
            Assert.AreEqual(typeof(ISupportBaseAB), selectTestCase.EventType.GetPropertyType("t2"));
            Assert.AreEqual(typeof(ISupportBaseABImpl), selectTestCase.EventType.GetPropertyType("t3"));
            Assert.AreEqual(typeof(ISupportC), selectTestCase.EventType.GetPropertyType("t4"));
            Assert.AreEqual(typeof(ISupportD), selectTestCase.EventType.GetPropertyType("t5"));
            Assert.AreEqual(typeof(ISupportAImplSuperG), selectTestCase.EventType.GetPropertyType("t6"));
            Assert.AreEqual(typeof(ISupportAImplSuperGImplPlus), selectTestCase.EventType.GetPropertyType("t7"));
    
            Object bean = new SupportBeanDynRoot("abc");
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(bean));
            EventBean theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {bean, null, null, null, null, null, null, null});
    
            bean = new ISupportDImpl("", "", "");
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(bean));
            theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {null, null, null, null, null, bean, null, null});
    
            bean = new ISupportBCImpl("", "", "");
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(bean));
            theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {null, null, bean, null, bean, null, null, null});
    
            bean = new ISupportAImplSuperGImplPlus();
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(bean));
            theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {null, bean, bean, null, bean, null, bean, bean});
    
            bean = new ISupportBaseABImpl("");
            _epService.EPRuntime.SendEvent(new SupportBeanDynRoot(bean));
            theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {null, null, bean, bean, null, null, null, null});
        }
    
        [Test]
        public void TestCastBoolean()
        {
            String stmtText = "select cast(BoolPrimitive as " + typeof(bool).FullName + ") as t0, " +
                              " cast(BoolBoxed | BoolPrimitive, boolean) as t1, " +
                              " cast(BoolBoxed, string) as t2 " +
                              " from " + typeof(SupportBean).FullName;
    
            EPStatement selectTestCase = _epService.EPAdministrator.CreateEPL(stmtText);
            selectTestCase.Events += _listener.Update;
    
            Assert.AreEqual(typeof(bool?), selectTestCase.EventType.GetPropertyType("t0"));
            Assert.AreEqual(typeof(bool?), selectTestCase.EventType.GetPropertyType("t1"));
            Assert.AreEqual(typeof(string), selectTestCase.EventType.GetPropertyType("t2"));
    
            SupportBean bean = new SupportBean("abc", 100);
            bean.BoolPrimitive = true;
            bean.BoolBoxed = true;
            _epService.EPRuntime.SendEvent(bean);
            EventBean theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {true, true, "True"});
    
            bean = new SupportBean(null, 100);
            bean.BoolPrimitive = false;
            bean.BoolBoxed = false;
            _epService.EPRuntime.SendEvent(bean);
            theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {false, false, "False"});
    
            bean = new SupportBean(null, 100);
            bean.BoolPrimitive = true;
            bean.BoolBoxed = null;
            _epService.EPRuntime.SendEvent(bean);
            theEvent = _listener.AssertOneGetNewAndReset();
            AssertResults(theEvent, new Object[] {true, null, null});
        }
    
        [Test]
        public void TestCastMapStringInt()
        {
            var map = new Dictionary<string,object>();
            map.Put("anInt",typeof(string));
            map.Put("anDouble",typeof(string));
            map.Put("anLong",typeof(string));
            map.Put("anFloat",typeof(string));
            map.Put("anByte",typeof(string));
            map.Put("anShort",typeof(string));
            map.Put("IntPrimitive",typeof(int));
            map.Put("IntBoxed",typeof(int));
    
            Configuration config = new Configuration();
            config.AddEventType("TestEvent", map);
    
            _epService = EPServiceProviderManager.GetDefaultProvider(config);
            _epService.Initialize();
    
            String stmt = "select cast(anInt, int) as intVal, " +
                                "cast(anDouble, double) as doubleVal, " +
                                "cast(anLong, long) as longVal, " +
                                "cast(anFloat, float) as floatVal, " +
                                "cast(anByte, byte) as byteVal, " +
                                "cast(anShort, short) as shortVal, " +
                                "cast(IntPrimitive, int) as intOne, " +
                                "cast(IntBoxed, int) as intTwo, " +
                                "cast(IntPrimitive, " + typeof(long).FullName + ") as longOne, " +
                                "cast(IntBoxed, long) as longTwo " +
                        "from TestEvent";
            
            EPStatement statement = _epService.EPAdministrator.CreateEPL(stmt);
            statement.Events += _listener.Update;
            
            map = new Dictionary<string,object>();
            map.Put("anInt","100");
            map.Put("anDouble","1.4E-1");
            map.Put("anLong","-10");
            map.Put("anFloat","1.001");
            map.Put("anByte","0x0A");
            map.Put("anShort","223");
            map.Put("IntPrimitive",10);
            map.Put("IntBoxed",11);
    
            _epService.EPRuntime.SendEvent(map, "TestEvent");
            EventBean row = _listener.AssertOneGetNewAndReset();
            Assert.AreEqual(100, row.Get("intVal"));
            Assert.AreEqual(0.14d, row.Get("doubleVal"));
            Assert.AreEqual(-10L, row.Get("longVal"));
            Assert.AreEqual(1.001f, row.Get("floatVal"));
            Assert.AreEqual((byte)10, row.Get("byteVal"));
            Assert.AreEqual((short)223, row.Get("shortVal"));
            Assert.AreEqual(10, row.Get("intOne"));
            Assert.AreEqual(11, row.Get("intTwo"));
            Assert.AreEqual(10L, row.Get("longOne"));
            Assert.AreEqual(11L, row.Get("longTwo"));
        }
    
        private void AssertResults(EventBean theEvent, Object[] result)
        {
            for (int i = 0; i < result.Length; i++)
            {
                Assert.AreEqual(result[i], theEvent.Get("t" + i), "failed for index " + i);
            }
        }
    }
}
