///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2013 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Threading;

using com.espertech.esper.client;
using com.espertech.esper.client.scopetest;
using com.espertech.esper.compat;
using com.espertech.esper.compat.collections;
using com.espertech.esper.compat.logging;
using com.espertech.esper.core.service;
using com.espertech.esper.filter;
using com.espertech.esper.metrics.instrumentation;
using com.espertech.esper.support.bean;
using com.espertech.esper.support.client;

using NUnit.Framework;

namespace com.espertech.esper.regression.view
{
    [TestFixture]
    public class TestFilterExpressionsOptimizable 
    {
        private static readonly ILog Log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
    
        private EPServiceProvider _epService;
        private SupportUpdateListener _listener;

        [SetUp]
        public void SetUp()
        {
            _listener = new SupportUpdateListener();
    
            var config = SupportConfigFactory.GetConfiguration();
            config.AddEventType("SupportEvent", typeof(SupportTradeEvent));
            config.AddEventType("SupportBean", typeof(SupportBean));
    
            _epService = EPServiceProviderManager.GetDefaultProvider(config);
            _epService.Initialize();
            if (InstrumentationHelper.ENABLED) { InstrumentationHelper.StartTest(_epService, GetType(), GetType().FullName); }
        }
    
        [TearDown]
        public void TearDown()
        {
            if (InstrumentationHelper.ENABLED) { InstrumentationHelper.EndTest(); }
            _listener = null;
        }
    
        [Test]
        public void TestOptimizablePerf()
        {
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("libSplit", typeof(MyLib).FullName, "LibSplit", FilterOptimizable.ENABLED);
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("libE1True", typeof(MyLib).FullName, "LibE1True", FilterOptimizable.ENABLED);
    
            // create listeners
            const int count = 10;
            var listeners = new SupportUpdateListener[count];
            for (var i = 0; i < count; i++) {
                listeners[i] = new SupportUpdateListener();
            }
    
            // Func(...) = value
            RunAssertionEquals("select * from SupportBean(libSplit(TheString) = !NUM!)", listeners);
    
            // Func(...) implied true
            RunAssertionBoolean("select * from SupportBean(libE1True(TheString))");
    
            // declared expression (...) = value
            _epService.EPAdministrator.CreateEPL("create expression thesplit {TheString => libSplit(TheString)}");
            RunAssertionEquals("select * from SupportBean(thesplit(*) = !NUM!)", listeners);
    
            // declared expression (...) implied true
            _epService.EPAdministrator.CreateEPL("create expression theE1Test {TheString => libE1True(TheString)}");
            RunAssertionBoolean("select * from SupportBean(theE1Test(*))");
        }
    
        [Test]
        public void TestOptimizableInspectFilter()
        {
            String epl;
    
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("funcOne", typeof(MyLib).FullName, "LibSplit", FilterOptimizable.DISABLED);
            epl = "select * from SupportBean(funcOne(TheString) = 0)";
            AssertFilterSingle(epl, FilterSpecCompiler.PROPERTY_NAME_BOOLEAN_EXPRESSION, FilterOperator.BOOLEAN_EXPRESSION);

            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("funcOneWDefault", typeof(MyLib).FullName, "LibSplit");
            epl = "select * from SupportBean(funcOneWDefault(TheString) = 0)";
            AssertFilterSingle(epl, "funcOneWDefault(TheString)", FilterOperator.EQUAL);

            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("funcTwo", typeof(MyLib).FullName, "LibSplit", FilterOptimizable.ENABLED);
            epl = "select * from SupportBean(funcTwo(TheString) = 0)";
            AssertFilterSingle(epl, "funcTwo(TheString)", FilterOperator.EQUAL);
            
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("libE1True", typeof(MyLib).FullName, "LibE1True", FilterOptimizable.ENABLED);
            epl = "select * from SupportBean(libE1True(TheString))";
            AssertFilterSingle(epl, "libE1True(TheString)", FilterOperator.EQUAL);
    
            epl = "select * from SupportBean(funcTwo( TheString ) > 10)";
            AssertFilterSingle(epl, "funcTwo(TheString)", FilterOperator.GREATER);
    
            _epService.EPAdministrator.CreateEPL("create expression thesplit {TheString => FuncOne(TheString)}");
    
            epl = "select * from SupportBean(thesplit(*) = 0)";
            AssertFilterSingle(epl, "thesplit(*)", FilterOperator.EQUAL);
    
            epl = "select * from SupportBean(libE1True(TheString))";
            AssertFilterSingle(epl, "libE1True(TheString)", FilterOperator.EQUAL);
    
            epl = "select * from SupportBean(thesplit(*) > 10)";
            AssertFilterSingle(epl, "thesplit(*)", FilterOperator.GREATER);

            epl = "expression housenumber alias for {10} select * from SupportBean(IntPrimitive = housenumber)";
            AssertFilterSingle(epl, "IntPrimitive", FilterOperator.EQUAL);

            epl = "expression housenumber alias for {IntPrimitive*10} select * from SupportBean(IntPrimitive = housenumber)";
            AssertFilterSingle(epl, ".boolean_expression", FilterOperator.BOOLEAN_EXPRESSION);
        }

        [Test]
        public void TestOrToInRewrite()
        {
            // test 'or' rewrite
            String[] filtersAB = new String[] {
                    "TheString = 'a' or TheString = 'b'",
                    "TheString = 'a' or 'b' = TheString",
                    "'a' = TheString or 'b' = TheString",
                    "'a' = TheString or TheString = 'b'",
            };
            foreach (String filter in filtersAB) {
                String epl = "select * from SupportBean(" + filter + ")";
                AssertFilterSingle(epl, "TheString", FilterOperator.IN_LIST_OF_VALUES);
                _epService.EPAdministrator.CreateEPL(epl).Events += _listener.Update;

                _epService.EPRuntime.SendEvent(new SupportBean("a", 0));
                Assert.IsTrue(_listener.GetAndClearIsInvoked());
                _epService.EPRuntime.SendEvent(new SupportBean("b", 0));
                Assert.IsTrue(_listener.GetAndClearIsInvoked());
                _epService.EPRuntime.SendEvent(new SupportBean("c", 0));
                Assert.IsFalse(_listener.GetAndClearIsInvoked());

                _epService.EPAdministrator.DestroyAllStatements();
            }

            // test no-rewrite
            String[] filtersNoIn = new String[] {
                    "IntPrimitive = 1 or TheString = 'a'",
                    "TheString is 'x' or TheString = 'a'",
                    "TheString != 'x' or TheString = 'a'",
                    "TheString = 'a' or TheString > 'b'",
                    "TheString = TheString or TheString = 'b'",
                    "TheString = 'a' or TheString = TheString"
            };
            foreach (String filter in filtersNoIn) {
                String epl = "select * from SupportBean(" + filter + ")";
                AssertFilterSingle(epl, ".boolean_expression", FilterOperator.BOOLEAN_EXPRESSION);
            }

            String eplX = "select * from SupportBean(IntPrimitive = 1 and (TheString='a' or TheString='b'))";
            AssertFilterTwo(eplX, "IntPrimitive", FilterOperator.EQUAL, "TheString", FilterOperator.IN_LIST_OF_VALUES);
        }
    
        private void RunAssertionEquals(String epl, SupportUpdateListener[] listeners)
        {
            // test function returns lookup value and "equals"
            for (var i = 0; i < listeners.Length; i++) {
                var stmt = _epService.EPAdministrator.CreateEPL(epl.Replace("!NUM!", Convert.ToString(i)));
                stmt.Events += listeners[i].Update;
            }

            MyLib.ResetCountInvoked();

            var loops = 1000;
            var delta = PerformanceObserver.TimeMillis(
                () =>
                {
                    for (var i = 0; i < loops; i++)
                    {
                        _epService.EPRuntime.SendEvent(new SupportBean("E_" + i%listeners.Length, 0));
                        var listener = listeners[i%listeners.Length];
                        Assert.IsTrue(listener.GetAndClearIsInvoked());
                    }
                });

            Assert.AreEqual(loops, MyLib.CountInvoked);
    
            Log.Info("Equals delta=" + delta);
            Assert.That(delta, Is.LessThan(1000), "Delta is " + delta);
            _epService.EPAdministrator.DestroyAllStatements();
        }
    
        private void RunAssertionBoolean(String epl)
        {
            // test function returns lookup value and "equals"
            var count = 10;
            for (var i = 0; i < count; i++) {
                var stmt = _epService.EPAdministrator.CreateEPL(epl);
                stmt.Events += _listener.Update;
            }
    
            MyLib.ResetCountInvoked();

            Assert.That(MyLib.CountInvoked, Is.EqualTo(0));

            var loops = 10000;
            var delta = PerformanceObserver.TimeMillis(
                () =>
                {
                    for (var i = 0; i < loops; i++)
                    {
                        var key = "E_" + i%100;
                        _epService.EPRuntime.SendEvent(new SupportBean(key, 0));
                        if (key.Equals("E_1"))
                        {
                            Assert.AreEqual(count, _listener.NewDataList.Count);
                            _listener.Reset();
                        }
                        else
                        {
                            Assert.IsFalse(_listener.IsInvoked);
                        }
                    }
                });

            // As noted in my analysis, the invocation count can actually be larger than the number of loops,
            // this occurs because of the way in which the weak references are handled in the caching layers
            // below.  Java soft-references behave much better, but we don't have access to them in the CLR.
            Assert.That(MyLib.CountInvoked, Is.GreaterThanOrEqualTo(loops), string.Format("MyLib.CountInvoked = {0}", MyLib.CountInvoked));
    
            Log.Info("Boolean delta=" + delta);
            Assert.That(delta, Is.LessThan(1000), "Delta is " + delta);
            _epService.EPAdministrator.DestroyAllStatements();
        }
    
        private void AssertFilterSingle(String epl, String expression, FilterOperator op)
        {
            var statementSPI = (EPStatementSPI) _epService.EPAdministrator.CreateEPL(epl);
            var param = GetFilterSingle(statementSPI);
            Assert.AreEqual(op, param.FilterOperator);
            Assert.AreEqual(expression, param.Lookupable.Expression);
        }

        private void AssertFilterTwo(String epl, String expressionOne, FilterOperator opOne, String expressionTwo, FilterOperator opTwo)
        {
            EPStatementSPI statementSPI = (EPStatementSPI)_epService.EPAdministrator.CreateEPL(epl);
            FilterValueSetParam[] multi = GetFilterMulti(statementSPI);
            Assert.AreEqual(2, multi.Length);
            Assert.AreEqual(opOne, multi[0].FilterOperator);
            Assert.AreEqual(expressionOne, multi[0].Lookupable.Expression);
            Assert.AreEqual(opTwo, multi[1].FilterOperator);
            Assert.AreEqual(expressionTwo, multi[1].Lookupable.Expression);
        }
    
        private FilterValueSetParam GetFilterSingle(EPStatementSPI statementSPI)
        {
            FilterValueSetParam[] @params = GetFilterMulti(statementSPI);
            Assert.AreEqual(1, @params.Length);
            return @params[0];
        }

        private FilterValueSetParam[] GetFilterMulti(EPStatementSPI statementSPI)
        {
            var filterServiceSPI = (FilterServiceSPI) statementSPI.StatementContext.FilterService;
            var set = filterServiceSPI.Take(Collections.SingletonList(statementSPI.StatementId));
            Assert.AreEqual(1, set.Filters.Count);
            var valueSet = set.Filters[0].FilterValueSet;
            return valueSet.Parameters;
        }
    
        public class MyLib
        {
            private static long _countInvoked;

            public static int LibSplit(String theString)
            {
                var key = theString.Split('_');
                Interlocked.Increment(ref _countInvoked);
                return int.Parse(key[1]);
            }
    
            public static bool LibE1True(String theString)
            {
                Interlocked.Increment(ref _countInvoked);
                return theString.Equals("E_1");
            }

            public static long CountInvoked
            {
                get { return Interlocked.Read(ref _countInvoked); }
            }

            public static void ResetCountInvoked()
            {
                Interlocked.Exchange(ref _countInvoked, 0);
            }
        }
    }
}
