///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2012 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Linq;

using com.espertech.esper.client;
using com.espertech.esper.client.scopetest;
using com.espertech.esper.compat;
using com.espertech.esper.compat.logging;
using com.espertech.esper.core.service;
using com.espertech.esper.filter;
using com.espertech.esper.support.bean;
using com.espertech.esper.support.client;

using NUnit.Framework;

namespace com.espertech.esper.regression.view
{
    [TestFixture]
    public class TestFilterExpressionsPluginSingleRow 
    {
        private static readonly ILog Log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
    
        private EPServiceProvider _epService;
        private SupportUpdateListener _listener;
    
        [SetUp]
        public void SetUp()
        {
            _listener = new SupportUpdateListener();
    
            var config = SupportConfigFactory.GetConfiguration();
            config.AddEventType("SupportEvent", typeof(SupportTradeEvent));
            config.AddEventType("SupportBean", typeof(SupportBean));
    
            _epService = EPServiceProviderManager.GetDefaultProvider(config);
            _epService.Initialize();
        }
    
        [TearDown]
        public void TearDown()
        {
            _listener = null;
        }
    
        [Test]
        public void TestPluginSingleRowFunctionFilterPerf() {
    
            // create listeners
            var count = 100;
            var listeners = new SupportUpdateListener[count];
            for (var i = 0; i < count; i++) {
                listeners[i] = new SupportUpdateListener();
            }
    
            // Func(...) = value
            RunAssertionEquals(listeners);
            _epService.EPAdministrator.DestroyAllStatements();
    
            // Func(...) implied true
            RunAssertionBoolean();
            _epService.EPAdministrator.DestroyAllStatements();
        }
    
        private void RunAssertionEquals(SupportUpdateListener[] listeners) {
    
            // test function returns lookup value and "equals"
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("libSplit", typeof(MyLib).FullName, "LibSplit", FilterOptimizable.ENABLED);
            for (var i = 0; i < listeners.Length; i++) {
                var stmt = _epService.EPAdministrator.CreateEPL("select * from SupportBean(libSplit(TheString) = " + i + ")");
                stmt.Events += listeners[i].Update;
            }
    
            var startTime = PerformanceObserver.MilliTime;
            MyLib.ResetCountInvoked();
            var loops = 1000;
            for (var i = 0; i < loops; i++) {
                _epService.EPRuntime.SendEvent(new SupportBean("E_" + i % 100, 0));
                var listener = listeners[i % 100];
                Assert.IsTrue(listener.GetAndClearIsInvoked());
            }
            var delta = PerformanceObserver.MilliTime - startTime;
            Assert.AreEqual(loops, MyLib.CountInvoked);
    
            Log.Info("Equals delta=" + delta);
            Assert.IsTrue(delta < 1000, "Delta is " + delta);
        }
    
        private void RunAssertionBoolean() {
    
            // test function returns lookup value and "equals"
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("libE1True", typeof(MyLib).FullName, "LibE1True", FilterOptimizable.ENABLED);
            var count = 100;
            for (var i = 0; i < count; i++) {
                var stmt = _epService.EPAdministrator.CreateEPL("select * from SupportBean(libE1True(TheString))");
                stmt.Events += _listener.Update;
            }
    
            var startTime = PerformanceObserver.MilliTime;
            MyLib.ResetCountInvoked();
            var loops = 10000;
            for (var i = 0; i < loops; i++) {
                var key = "E_" + i % 100;
                _epService.EPRuntime.SendEvent(new SupportBean(key, 0));
                if (key.Equals("E_1")) {
                    Assert.AreEqual(count, _listener.NewDataList.Count);
                    _listener.Reset();
                }
                else {
                    Assert.IsFalse(_listener.IsInvoked);
                }
            }
            var delta = PerformanceObserver.MilliTime - startTime;
            Assert.AreEqual(loops, MyLib.CountInvoked);
    
            Log.Info("Boolea delta=" + delta);
            Assert.IsTrue(delta < 1000, "Delta is " + delta);
        }
    
        [Test]
        public void TestPluginSingleRowFunctionFilter() {
    
            String epl;
    
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("funcOne", typeof(MyLib).FullName, "LibSplit", FilterOptimizable.DISABLED);
            epl = "select * from SupportBean(funcOne(TheString) = 0)";
            AssertFilterSingle(epl, FilterSpecCompiler.PROPERTY_NAME_BOOLEAN_EXPRESSION, FilterOperator.BOOLEAN_EXPRESSION);

            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("funcOneWDefault", typeof(MyLib).FullName, "LibSplit");
            epl = "select * from SupportBean(funcOneWDefault(TheString) = 0)";
            AssertFilterSingle(epl, "funcOneWDefault(TheString)", FilterOperator.EQUAL);

            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("funcTwo", typeof(MyLib).FullName, "LibSplit", FilterOptimizable.ENABLED);
            epl = "select * from SupportBean(funcTwo(TheString) = 0)";
            AssertFilterSingle(epl, "funcTwo(TheString)", FilterOperator.EQUAL);
            
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction("libE1True", typeof(MyLib).FullName, "LibE1True", FilterOptimizable.ENABLED);
            epl = "select * from SupportBean(libE1True(TheString))";
            AssertFilterSingle(epl, "libE1True(TheString)", FilterOperator.EQUAL);

            epl = "select * from SupportBean(funcTwo(TheString) > 10)";
            AssertFilterSingle(epl, "funcTwo(TheString)", FilterOperator.GREATER);
        }
    
        private void AssertFilterSingle(String epl, String expression, FilterOperator op)
        {
            var statementSPI = (EPStatementSPI) _epService.EPAdministrator.CreateEPL(epl);
            var param = GetFilterSingle(statementSPI);
            Assert.AreEqual(op, param.FilterOperator);
            Assert.AreEqual(expression, param.Lookupable.Expression);
        }
    
        private FilterValueSetParam GetFilterSingle(EPStatementSPI statementSPI)
        {
            var filterServiceSPI = (FilterServiceSPI) statementSPI.StatementContext.FilterService;
            var set = filterServiceSPI.Take(statementSPI.StatementId.AsSingleton());
            Assert.AreEqual(1, set.Filters.Count);
            var valueSet = set.Filters[0].FilterValueSet;
            Assert.AreEqual(1, valueSet.Parameters.Length);
            return valueSet.Parameters.First();
        }
    
        public class MyLib
        {
            private static int _countInvoked;
    
            public static int LibSplit(String stringValue)
            {
                var key = stringValue.Split('_');
                _countInvoked++;
                return int.Parse(key[1]);
            }
    
            public static bool LibE1True(String stringValue) {
                _countInvoked++;
                return stringValue.Equals("E_1");
            }

            public static int CountInvoked
            {
                get { return _countInvoked; }
            }

            public static void ResetCountInvoked() {
                _countInvoked = 0;
            }
        }
    }
}
