///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2012 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;

using com.espertech.esper.client;
using com.espertech.esper.client.scopetest;
using com.espertech.esper.client.soda;
using com.espertech.esper.compat.collections;
using com.espertech.esper.metrics.instrumentation;
using com.espertech.esper.support.bean;
using com.espertech.esper.support.client;
using com.espertech.esper.util;

using NUnit.Framework;

namespace com.espertech.esper.regression.view
{
    using Map = IDictionary<string, object>;

    [TestFixture]
    public class TestFilterPropertySplitExpr
    {
        private EPServiceProvider _epService;
        private SupportUpdateListener _listener;

        [SetUp]
        public void SetUp()
        {
            _epService = EPServiceProviderManager.GetDefaultProvider(
                SupportConfigFactory.GetConfiguration());
            _epService.Initialize();
            if (InstrumentationHelper.ENABLED) { InstrumentationHelper.StartTest(_epService, GetType(), GetType().FullName); }
            _listener = new SupportUpdateListener();
        }

        [TearDown]
        public void TearDown()
        {
            if (InstrumentationHelper.ENABLED) { InstrumentationHelper.EndTest(); }
            _listener = null;
        }

        private void RunAssertionSingleRowSplitAndType(EventRepresentationEnum eventRepresentationEnum)
        {
            if (eventRepresentationEnum.IsObjectArrayEvent())
            {
                _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction(
                    "splitSentence", GetType().FullName,
                    "SplitSentenceMethodReturnObjectArray");
                _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction(
                    "splitSentenceBean", GetType().FullName,
                    "SplitSentenceBeanMethodReturnObjectArray");
                _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction(
                    "splitWord", GetType().FullName,
                    "SplitWordMethodReturnObjectArray");
            }
            else
            {
                _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction(
                    "splitSentence", GetType().FullName,
                    "SplitSentenceMethodReturnMap");
                _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction(
                    "splitSentenceBean", GetType().FullName,
                    "SplitSentenceBeanMethodReturnMap");
                _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction(
                    "splitWord", GetType().FullName,
                    "SplitWordMethodReturnMap");
            }
            _epService.EPAdministrator.Configuration.AddPlugInSingleRowFunction(
                "invalidSentence", GetType().FullName,
                "InvalidSentenceMethod");

            _epService.EPAdministrator.CreateEPL(
                eventRepresentationEnum.GetAnnotationText()
                + " create schema SentenceEvent(sentence String)");
            _epService.EPAdministrator.CreateEPL(
                eventRepresentationEnum.GetAnnotationText()
                + " create schema WordEvent(word String)");
            _epService.EPAdministrator.CreateEPL(
                eventRepresentationEnum.GetAnnotationText()
                + " create schema CharacterEvent(char String)");

            String stmtText;
            EPStatement stmt;
            String[] fields = "word".Split(',');

            // test single-row method
            stmtText = "select * from SentenceEvent[splitSentence(sentence)@type(WordEvent)]";
            stmt = _epService.EPAdministrator.CreateEPL(stmtText);
            stmt.Events += _listener.Update;
            Assert.AreEqual("WordEvent", stmt.EventType.Name);
            Assert.AreEqual(eventRepresentationEnum.GetOutputClass(),
                            stmt.EventType.UnderlyingType);

            SendSentenceEvent(eventRepresentationEnum, "I am testing this code");
            EPAssertionUtil.AssertPropsPerRow(
                _listener.GetAndResetLastNewData(),
                fields, new Object[][]
                {
                    new Object[] { "I" },
                    new Object[] { "am" },
                    new Object[] { "testing" },
                    new Object[] { "this" },
                    new Object[] { "code" }
                });

            SendSentenceEvent(eventRepresentationEnum, "the second event");
            EPAssertionUtil.AssertPropsPerRow(
                _listener.GetAndResetLastNewData(),
                fields, new Object[][]
                {
                    new Object[] { "the" },
                    new Object[] { "second" },
                    new Object[] { "event" }
                });

            stmt.Dispose();

            // test SODA
            EPStatementObjectModel model = _epService.EPAdministrator.CompileEPL(
                stmtText);

            Assert.AreEqual(stmtText, model.ToEPL());
            stmt = _epService.EPAdministrator.Create(model);
            Assert.AreEqual(stmtText, stmt.Text);
            stmt.Events += _listener.Update;

            SendSentenceEvent(eventRepresentationEnum, "the third event");
            EPAssertionUtil.AssertPropsPerRow(
                _listener.GetAndResetLastNewData(),
                fields, new Object[][]
                {
                    new Object[] { "the" },
                    new Object[] { "third" },
                    new Object[] { "event" }
                });

            stmt.Dispose();

#if DISABLED
            // test script
            if (!eventRepresentationEnum.IsObjectArrayEvent())
            {
                stmtText = "expression Collection js:splitSentenceJS(sentence) ["
                           + "  ImportPackage(java.util);"
                           + "  var words = new ArrayList();"
                           + "  words.Add(Collections.SingletonMap('word', 'wordOne'));"
                           + "  words.Add(Collections.SingletonMap('word', 'wordTwo'));"
                           + "  words;" + "]"
                           + "select * from SentenceEvent[splitSentenceJS(sentence)@type(WordEvent)]";
                stmt = _epService.EPAdministrator.CreateEPL(stmtText);
                stmt.Events += _listener.Update;
                Assert.AreEqual("WordEvent", stmt.EventType.Name);

                _epService.EPRuntime.SendEvent(Collections.EmptyDataMap, "SentenceEvent");
                EPAssertionUtil.AssertPropsPerRowAnyOrder(
                    _listener.GetAndResetLastNewData(),
                    fields, new Object[][]
                    {
                        new Object[] { "wordOne" },
                        new Object[] { "wordTwo" }
                    });

                stmt.Dispose();
            }
#endif

            // test multiple splitters
            stmtText =
                "select * from SentenceEvent[splitSentence(sentence)@type(WordEvent)][splitWord(word)@type(CharacterEvent)]";
            stmt = _epService.EPAdministrator.CreateEPL(stmtText);
            stmt.Events += _listener.Update;
            Assert.AreEqual("CharacterEvent", stmt.EventType.Name);

            SendSentenceEvent(eventRepresentationEnum, "I am");
            EPAssertionUtil.AssertPropsPerRowAnyOrder(
                _listener.GetAndResetLastNewData(),
                "char".Split(','), new Object[][]
                {
                    new Object[] { "I" },
                    new Object[] { "a" },
                    new Object[] { "m" }
                });

            stmt.Dispose();

            // test wildcard parameter
            stmtText = "select * from SentenceEvent[splitSentenceBean(*)@type(WordEvent)]";
            stmt = _epService.EPAdministrator.CreateEPL(stmtText);
            stmt.Events += _listener.Update;
            Assert.AreEqual("WordEvent", stmt.EventType.Name);

            SendSentenceEvent(eventRepresentationEnum, "another test sentence");
            EPAssertionUtil.AssertPropsPerRowAnyOrder(
                _listener.GetAndResetLastNewData(),
                fields, new Object[][]
                {
                    new Object[] { "another" },
                    new Object[] { "test" },
                    new Object[] { "sentence" }
                });

            stmt.Dispose();

            // test property returning untyped collection
            if (eventRepresentationEnum.IsObjectArrayEvent())
            {
                _epService.EPAdministrator.Configuration.AddEventType(
                    typeof (ObjectArrayEvent));
                stmtText = eventRepresentationEnum.GetAnnotationText()
                           + " select * from ObjectArrayEvent[someObjectArray@type(WordEvent)]";
                stmt = _epService.EPAdministrator.CreateEPL(stmtText);
                stmt.Events += _listener.Update;
                Assert.AreEqual("WordEvent", stmt.EventType.Name);

                var rows = new Object[][]
                {
                    new Object[] { "this" },
                    new Object[] { "is" },
                    new Object[] { "collection" }
                };

                _epService.EPRuntime.SendEvent(new ObjectArrayEvent(rows));
                EPAssertionUtil.AssertPropsPerRow(
                    _listener.GetAndResetLastNewData(),
                    fields, new Object[][]
                    {
                        new Object[] { "this" },
                        new Object[] { "is" },
                        new Object[] { "collection" }
                    });
            }
            else
            {
                _epService.EPAdministrator.Configuration.AddEventType<CollectionEvent>();
                stmtText = eventRepresentationEnum.GetAnnotationText()
                           + " select * from CollectionEvent[someCollection@type(WordEvent)]";
                stmt = _epService.EPAdministrator.CreateEPL(stmtText);
                stmt.Events += _listener.Update;
                Assert.AreEqual("WordEvent", stmt.EventType.Name);

                ICollection<Map> coll = new List<Map>();

                coll.Add(Collections.SingletonMap<string, object>("word", "this"));
                coll.Add(Collections.SingletonMap<string, object>("word", "is"));
                coll.Add(Collections.SingletonMap<string, object>("word", "collection"));

                _epService.EPRuntime.SendEvent(new CollectionEvent(coll));
                EPAssertionUtil.AssertPropsPerRowAnyOrder(
                    _listener.GetAndResetLastNewData(),
                    fields, new Object[][]
                    {
                        new Object[] { "this" },
                        new Object[] { "is" },
                        new Object[] { "collection" }
                    });
            }

            // invalid: event type not found
            TryInvalid(
                "select * from SentenceEvent[splitSentence(sentence)@type(XYZ)]",
                "Event type by name 'XYZ' could not be found [select * from SentenceEvent[splitSentence(sentence)@type(XYZ)]]");

            // invalid lib-function annotation
            TryInvalid(
                "select * from SentenceEvent[splitSentence(sentence)@dummy(WordEvent)]",
                "Invalid annotation for property selection, expected 'type' but found 'dummy' in text '[splitSentence(sentence)@dummy(WordEvent)]' [select * from SentenceEvent[splitSentence(sentence)@dummy(WordEvent)]]");

            // invalid type assignment to event type
            if (eventRepresentationEnum.IsObjectArrayEvent())
            {
                TryInvalid(
                    "select * from SentenceEvent[invalidSentence(sentence)@type(WordEvent)]",
                    "Event type 'WordEvent' underlying type System.Object(Array) cannot be assigned a value of type");
            }
            else
            {
                TryInvalid(
                    "select * from SentenceEvent[invalidSentence(sentence)@type(WordEvent)]",
                    "Event type 'WordEvent' underlying type " + typeof(Map).GetTypeNameFullyQualPretty() + " cannot be assigned a value of type");
            }

            // invalid subquery
            TryInvalid(
                "select * from SentenceEvent[splitSentence((select * from SupportBean.win:keepall()))@type(WordEvent)]",
                "Invalid contained-event expression 'splitSentence(subselect_0)': Aggregation, sub-select, previous or prior functions are not supported in this context [select * from SentenceEvent[splitSentence((select * from SupportBean.win:keepall()))@type(WordEvent)]]");

            _epService.Initialize();
        }

        private void SendSentenceEvent(EventRepresentationEnum eventRepresentationEnum, String sentence)
        {
            if (eventRepresentationEnum.IsObjectArrayEvent())
            {
                _epService.EPRuntime.SendEvent(new Object[]
                {
                    sentence
                }, "SentenceEvent");
            }
            else
            {
                _epService.EPRuntime.SendEvent(
                    Collections.SingletonMap<string, object>("sentence", sentence),
                    "SentenceEvent");
            }
        }

        public static Map[] SplitSentenceMethodReturnMap(String sentence)
        {
            String[] words = sentence.Split(' ');
            var events = new Map[words.Length];

            for (int i = 0; i < words.Length; i++)
            {
                events[i] = Collections.SingletonMap<string, object>("word", words[i]);
            }
            return events;
        }

        public static Object[][] SplitSentenceMethodReturnObjectArray(String sentence)
        {
            String[] words = sentence.Split(' ');
            var events = new Object[words.Length][];

            for (int i = 0; i < words.Length; i++)
            {
                events[i] = new Object[]
                {
                    words[i]
                };
            }
            return events;
        }

        public static Map[] SplitSentenceBeanMethodReturnMap(Map sentenceEvent)
        {
            return SplitSentenceMethodReturnMap(
                (String) sentenceEvent.Get("sentence"));
        }

        public static Object[][] SplitSentenceBeanMethodReturnObjectArray(Object[] sentenceEvent)
        {
            return SplitSentenceMethodReturnObjectArray((String) sentenceEvent[0]);
        }

        public static Object[][] SplitWordMethodReturnObjectArray(String word)
        {
            int count = word.Length;
            var events = new Object[count][];

            for (int i = 0; i < word.Length; i++)
            {
                events[i] = new Object[]
                {
                    word[i].ToString()
                };
            }
            return events;
        }

        public static Map[] SplitWordMethodReturnMap(String word)
        {
            var maps = new List<Map>();

            for (int i = 0; i < word.Length; i++)
            {
                maps.Add(Collections.SingletonMap<string, object>("char", word[i].ToString()));
            }
            return maps.ToArray();
        }

        public static SupportBean[] InvalidSentenceMethod(String sentence)
        {
            return null;
        }

        private void TryInvalid(String epl, String expected)
        {
            try
            {
                _epService.EPAdministrator.CreateEPL(epl);
                Assert.Fail();
            }
            catch (EPStatementException ex)
            {
                Assert.That(expected, Has.Length.GreaterThan(0));
                Assert.That(ex.Message, Is.StringStarting(expected));
            }
        }

        public class CollectionEvent
        {
            public CollectionEvent(ICollection<Map> someCollection)
            {
                SomeCollection = someCollection;
            }

            public ICollection<Map> SomeCollection { get; set; }
        }


        public class ObjectArrayEvent
        {
            public ObjectArrayEvent(Object[][] someObjectArray)
            {
                SomeObjectArray = someObjectArray;
            }

            public object[][] SomeObjectArray { get; private set; }
        }

        [Test]
        public void TestSingleRowSplitAndType()
        {
            RunAssertionSingleRowSplitAndType(EventRepresentationEnum.OBJECTARRAY);
            RunAssertionSingleRowSplitAndType(EventRepresentationEnum.MAP);
            RunAssertionSingleRowSplitAndType(EventRepresentationEnum.DEFAULT);
        }
    }
}
