///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2014 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;

using com.espertech.esper.client;
using com.espertech.esper.compat;
using com.espertech.esper.compat.collections;
using com.espertech.esper.epl.agg.access;
using com.espertech.esper.epl.agg.aggregator;
using com.espertech.esper.epl.expression.core;
using com.espertech.esper.epl.table.mgmt;
using com.espertech.esper.epl.table.strategy;
using com.espertech.esper.events;
using com.espertech.esper.metrics.instrumentation;

namespace com.espertech.esper.epl.agg.service
{
    /// <summary>
    /// Implementation for handling aggregation with grouping by group-keys.
    /// </summary>
    public class AggSvcGroupByWTableImpl : AggregationService
    {
        protected readonly TableMetadata tableMetadata;
        protected readonly TableColumnMethodPair[] methodPairs;
        protected readonly AggregationAccessorSlotPair[] accessors;
        protected readonly bool isJoin;
        protected readonly TableStateInstanceGroupBy tableStateInstance;
        protected readonly int[] targetStates;
        protected readonly ExprNode[] accessStateExpr;
        private readonly AggregationAgent[] agents;
    
        // maintain a current row for random access into the aggregator state table
        // (row=groups, columns=expression nodes that have aggregation functions)
        private AggregationMethod[] _currentAggregatorMethods;
        private AggregationState[] _currentAggregatorStates;
        private object _currentGroupKey;

        public AggSvcGroupByWTableImpl(
            TableMetadata tableMetadata,
            TableColumnMethodPair[] methodPairs,
            AggregationAccessorSlotPair[] accessors,
            bool join,
            TableStateInstanceGroupBy tableStateInstance,
            int[] targetStates,
            ExprNode[] accessStateExpr,
            AggregationAgent[] agents)
        {
            this.tableMetadata = tableMetadata;
            this.methodPairs = methodPairs;
            this.accessors = accessors;
            isJoin = join;
            this.tableStateInstance = tableStateInstance;
            this.targetStates = targetStates;
            this.accessStateExpr = accessStateExpr;
            this.agents = agents;
        }
    
        public void ClearResults(ExprEvaluatorContext exprEvaluatorContext)
        {
            tableStateInstance.Rows.Clear();
        }
    
        public void ApplyEnter(EventBean[] eventsPerStream, object groupByKey, ExprEvaluatorContext exprEvaluatorContext)
        {
            // acquire table-level write lock
            ExprTableEvalLockUtil.ObtainLockUnless(tableStateInstance.TableLevelRWLock.WriteLock, exprEvaluatorContext);
    
            var bean = tableStateInstance.GetCreateRowIntoTable(groupByKey, exprEvaluatorContext);
            var row = (AggregationRowPair) bean.Properties[0];
    
            _currentAggregatorMethods = row.Methods;
            _currentAggregatorStates = row.States;
    
            if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().QAggregationGroupedApplyEnterLeave(true, methodPairs.Length, targetStates.Length, groupByKey);}

            var evaluateParams = new EvaluateParams(eventsPerStream, true, exprEvaluatorContext);
            for (var j = 0; j < methodPairs.Length; j++)
            {
                var methodPair = methodPairs[j];
                var method = _currentAggregatorMethods[methodPair.TargetIndex];
                if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().QAggNoAccessEnterLeave(true, j, method, methodPair.AggregationNode);}
                var columnResult = methodPair.Evaluator.Evaluate(evaluateParams);
                method.Enter(columnResult);
                if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().AAggNoAccessEnterLeave(true, j, method);}
            }
    
            for (var i = 0; i < targetStates.Length; i++) {
                var state = _currentAggregatorStates[targetStates[i]];
                if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().QAggAccessEnterLeave(true, i, state, accessStateExpr[i]);}
                agents[i].ApplyEnter(eventsPerStream, exprEvaluatorContext, state);
                if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().AAggAccessEnterLeave(true, i, state);}
            }
    
            tableStateInstance.HandleRowUpdated(bean);
            if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().AAggregationGroupedApplyEnterLeave(true);}
        }
    
        public void ApplyLeave(EventBean[] eventsPerStream, object groupByKey, ExprEvaluatorContext exprEvaluatorContext)
        {
            // acquire table-level write lock
            ExprTableEvalLockUtil.ObtainLockUnless(tableStateInstance.TableLevelRWLock.WriteLock, exprEvaluatorContext);
    
            var bean = tableStateInstance.GetCreateRowIntoTable(groupByKey, exprEvaluatorContext);
            var row = (AggregationRowPair) bean.Properties[0];
            var evaluateParams = new EvaluateParams(eventsPerStream, false, exprEvaluatorContext);
    
            _currentAggregatorMethods = row.Methods;
            _currentAggregatorStates = row.States;
    
            if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().QAggregationGroupedApplyEnterLeave(false, methodPairs.Length, targetStates.Length, groupByKey);}
    
            for (var j = 0; j < methodPairs.Length; j++) {
                var methodPair = methodPairs[j];
                var method = _currentAggregatorMethods[methodPair.TargetIndex];
                if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().QAggNoAccessEnterLeave(false, j, method, methodPair.AggregationNode);}
                var columnResult = methodPair.Evaluator.Evaluate(evaluateParams);
                method.Leave(columnResult);
                if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().AAggNoAccessEnterLeave(false, j, method);}
            }
    
            for (var i = 0; i < targetStates.Length; i++) {
                var state = _currentAggregatorStates[targetStates[i]];
                if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().QAggAccessEnterLeave(false, i, state, accessStateExpr[i]);}
                agents[i].ApplyLeave(eventsPerStream, exprEvaluatorContext, state);
                if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().AAggAccessEnterLeave(false, i, state);}
            }
    
            tableStateInstance.HandleRowUpdated(bean);
            if (InstrumentationHelper.ENABLED) { InstrumentationHelper.Get().AAggregationGroupedApplyEnterLeave(false);}
        }
    
        public void SetCurrentAccess(object groupByKey, int agentInstanceId, AggregationGroupByRollupLevel rollupLevel)
        {
            var bean = tableStateInstance.Rows.Get(groupByKey);
            if (bean != null) {
                var row = (AggregationRowPair) bean.Properties[0];
                _currentAggregatorMethods = row.Methods;
                _currentAggregatorStates = row.States;
            }
            else {
                _currentAggregatorMethods = null;
            }
    
            this._currentGroupKey = groupByKey;
        }
    
        public object GetValue(int column, int agentInstanceId, EventBean[] eventsPerStream, bool isNewData, ExprEvaluatorContext exprEvaluatorContext) {
            if (column < _currentAggregatorMethods.Length) {
                return _currentAggregatorMethods[column].Value;
            }
            else {
                var pair = accessors[column - _currentAggregatorMethods.Length];
                return pair.Accessor.GetValue(_currentAggregatorStates[pair.Slot], eventsPerStream, isNewData, exprEvaluatorContext);
            }
        }
    
        public ICollection<EventBean> GetCollectionOfEvents(int column, EventBean[] eventsPerStream, bool isNewData, ExprEvaluatorContext context) {
            if (column < _currentAggregatorMethods.Length) {
                return null;
            }
            else {
                var pair = accessors[column - _currentAggregatorMethods.Length];
                return pair.Accessor.GetEnumerableEvents(_currentAggregatorStates[pair.Slot], eventsPerStream, isNewData, context);
            }
        }
    
        public ICollection<object> GetCollectionScalar(int column, EventBean[] eventsPerStream, bool isNewData, ExprEvaluatorContext context) {
            if (column < _currentAggregatorMethods.Length) {
                return null;
            }
            else {
                var pair = accessors[column - _currentAggregatorMethods.Length];
                return pair.Accessor.GetEnumerableScalar(_currentAggregatorStates[pair.Slot], eventsPerStream, isNewData, context);
            }
        }
    
        public EventBean GetEventBean(int column, EventBean[] eventsPerStream, bool isNewData, ExprEvaluatorContext context) {
            if (column < _currentAggregatorMethods.Length) {
                return null;
            }
            else {
                var pair = accessors[column - _currentAggregatorMethods.Length];
                return pair.Accessor.GetEnumerableEvent(_currentAggregatorStates[pair.Slot], eventsPerStream, isNewData, context);
            }
        }
    
        public void SetRemovedCallback(AggregationRowRemovedCallback callback) {
            // not applicable
        }
    
        public void Accept(AggregationServiceVisitor visitor) {
            // not applicable
        }
    
        public void AcceptGroupDetail(AggregationServiceVisitorWGroupDetail visitor) {
            // not applicable
        }

        public bool IsGrouped
        {
            get { return true; }
        }

        public object GetGroupKey(int agentInstanceId) {
            return _currentGroupKey;
        }
    
        public ICollection<object> GetGroupKeys(ExprEvaluatorContext exprEvaluatorContext) {
            return tableStateInstance.Rows.Keys;
        }
    }
}
