///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2014 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;

using com.espertech.esper.client;
using com.espertech.esper.client.annotation;
using com.espertech.esper.compat.collections;
using com.espertech.esper.epl.agg.access;
using com.espertech.esper.epl.core;
using com.espertech.esper.epl.declexpr;
using com.espertech.esper.epl.expression.baseagg;
using com.espertech.esper.epl.expression.core;
using com.espertech.esper.epl.expression.table;
using com.espertech.esper.epl.expression.visitor;
using com.espertech.esper.epl.spec;
using com.espertech.esper.epl.table.mgmt;
using com.espertech.esper.epl.util;
using com.espertech.esper.epl.variable;

namespace com.espertech.esper.epl.agg.service
{
	/// <summary>
	/// Factory for aggregation service instances.
	/// <para />Consolidates aggregation nodes such that result futures point to a single instance and
	/// no re-evaluation of the same result occurs.
	/// </summary>
	public class AggregationServiceFactoryFactory
	{
        /// <summary>
        /// Produces an aggregation service for use with match-recognice.
        /// </summary>
        /// <param name="numStreams">number of streams</param>
        /// <param name="measureExprNodesPerStream">measure nodes</param>
        /// <param name="typesPerStream">The types per stream.</param>
        /// <returns>
        /// service
        /// </returns>
	    public static AggregationServiceMatchRecognizeFactoryDesc GetServiceMatchRecognize(
	        int numStreams,
	        IDictionary<int, IList<ExprAggregateNode>> measureExprNodesPerStream,
	        EventType[] typesPerStream)
	    {
	        var equivalencyListPerStream = new SortedDictionary<int, IList<AggregationServiceAggExpressionDesc>>();

	        foreach (var entry in measureExprNodesPerStream)
	        {
	            IList<AggregationServiceAggExpressionDesc> equivalencyList = new List<AggregationServiceAggExpressionDesc>();
	            equivalencyListPerStream.Put(entry.Key, equivalencyList);
	            foreach (var selectAggNode in entry.Value)
	            {
	                AddEquivalent(selectAggNode, equivalencyList);
	            }
	        }

	        var aggregatorsPerStream = new LinkedHashMap<int, AggregationMethodFactory[]>();
	        var evaluatorsPerStream = new Dictionary<int, ExprEvaluator[]>();

	        foreach (var equivalencyPerStream in equivalencyListPerStream)
	        {
	            var index = 0;
	            int stream = equivalencyPerStream.Key;

	            var aggregators = new AggregationMethodFactory[equivalencyPerStream.Value.Count];
	            aggregatorsPerStream.Put(stream, aggregators);

	            var evaluators = new ExprEvaluator[equivalencyPerStream.Value.Count];
	            evaluatorsPerStream.Put(stream, evaluators);

	            foreach (var aggregation in equivalencyPerStream.Value)
	            {
	                var aggregateNode = aggregation.AggregationNode;
	                if (aggregateNode.ChildNodes.Length > 1)
	                {
	                    evaluators[index] = GetMultiNodeEvaluator(aggregateNode.ChildNodes, typesPerStream.Length > 1, typesPerStream);
	                }
	                else if (aggregateNode.ChildNodes.Length > 0)
	                {
	                    // Use the evaluation node under the aggregation node to obtain the aggregation value
	                    evaluators[index] = aggregateNode.ChildNodes[0].ExprEvaluator;
	                }
	                // For aggregation that doesn't evaluate any particular sub-expression, return null on evaluation
	                else
	                {
	                    evaluators[index] = new ProxyExprEvaluator
                        {
	                        ProcEvaluate = args => null,
	                        ProcReturnType = () => null
	                    };
	                }

	                aggregators[index] = aggregateNode.Factory;
	                index++;
	            }
	        }

	        // Assign a column number to each aggregation node. The regular aggregation goes first followed by access-aggregation.
	        var columnNumber = 0;
	        IList<AggregationServiceAggExpressionDesc> allExpressions = new List<AggregationServiceAggExpressionDesc>();
	        foreach (var equivalencyPerStream in equivalencyListPerStream)
	        {
	            foreach (var entry in equivalencyPerStream.Value)
	            {
	                entry.ColumnNum = columnNumber++;
	            }
	            allExpressions.AddAll(equivalencyPerStream.Value);
	        }

	        AggregationServiceMatchRecognizeFactory factory = new AggregationServiceMatchRecognizeFactoryImpl(numStreams, aggregatorsPerStream, evaluatorsPerStream);
	        return new AggregationServiceMatchRecognizeFactoryDesc(factory, allExpressions);
	    }

        /// <summary>
        /// Returns an instance to handle the aggregation required by the aggregation expression nodes, depending on
        /// whether there are any group-by nodes.
        /// </summary>
        /// <param name="selectAggregateExprNodes">aggregation nodes extracted out of the select expression</param>
        /// <param name="selectClauseNamedNodes">The select clause named nodes.</param>
        /// <param name="declaredExpressions">The declared expressions.</param>
        /// <param name="groupByNodes">The group by nodes.</param>
        /// <param name="havingAggregateExprNodes">aggregation nodes extracted out of the select expression</param>
        /// <param name="orderByAggregateExprNodes">aggregation nodes extracted out of the select expression</param>
        /// <param name="groupKeyExpressions">The group key expressions.</param>
        /// <param name="hasGroupByClause">indicator on whethere there is group-by required, or group-all</param>
        /// <param name="exprEvaluatorContext">context for expression evaluatiom</param>
        /// <param name="annotations">statement annotations</param>
        /// <param name="variableService">variable</param>
        /// <param name="isJoin">true for joins</param>
        /// <param name="isDisallowNoReclaim">if set to <c>true</c> [is disallow no reclaim].</param>
        /// <param name="whereClause">the where-clause function if any</param>
        /// <param name="havingClause">the having-clause function if any</param>
        /// <param name="factoryService">The factory service.</param>
        /// <param name="typesPerStream">The types per stream.</param>
        /// <param name="methodResolutionService">The method resolution service.</param>
        /// <param name="groupByRollupDesc">The group by rollup desc.</param>
        /// <param name="optionalContextName">Name of the optional context.</param>
        /// <param name="intoTableSpec">The into table spec.</param>
        /// <param name="tableService">The table service.</param>
        /// <returns>
        /// instance for aggregation handling
        /// </returns>
        /// <throws>com.espertech.esper.epl.expression.core.ExprValidationException if validation fails</throws>
	    public static AggregationServiceFactoryDesc GetService(
	        IList<ExprAggregateNode> selectAggregateExprNodes,
	        IDictionary<ExprNode, string> selectClauseNamedNodes,
	        IList<ExprDeclaredNode> declaredExpressions,
	        ExprNode[] groupByNodes,
	        IList<ExprAggregateNode> havingAggregateExprNodes,
	        IList<ExprAggregateNode> orderByAggregateExprNodes,
	        IList<ExprAggregateNodeGroupKey> groupKeyExpressions,
	        bool hasGroupByClause,
	        ExprEvaluatorContext exprEvaluatorContext,
	        Attribute[] annotations,
	        VariableService variableService,
	        bool isJoin,
	        bool isDisallowNoReclaim,
	        ExprNode whereClause,
	        ExprNode havingClause,
	        AggregationServiceFactoryService factoryService,
	        EventType[] typesPerStream,
	        MethodResolutionService methodResolutionService,
	        AggregationGroupByRollupDesc groupByRollupDesc,
	        string optionalContextName,
	        IntoTableSpec intoTableSpec,
	        TableService tableService)
	    {
	        // No aggregates used, we do not need this service
	        if ((selectAggregateExprNodes.IsEmpty()) && (havingAggregateExprNodes.IsEmpty()))
	        {
	            if (intoTableSpec != null) {
	                throw new ExprValidationException("Into-table requires at least one aggregation function");
	            }

	            return new AggregationServiceFactoryDesc(
	                factoryService.GetNullAggregationService(),
	                Collections.GetEmptyList<AggregationServiceAggExpressionDesc>(),
	                Collections.GetEmptyList<ExprAggregateNodeGroupKey>());
	        }

	        // Validate the absence of "prev" function in where-clause:
	        // Since the "previous" function does not post remove stream results, disallow when used with aggregations.
	        if ((whereClause != null) || (havingClause != null)) {
	            var visitor = new ExprNodePreviousVisitorWParent();
	            if (whereClause != null) {
	                whereClause.Accept(visitor);
	            }
	            if (havingClause != null) {
	                havingClause.Accept(visitor);
	            }
	            if ((visitor.Previous != null) && (!visitor.Previous.IsEmpty())) {
	                string funcname = visitor.Previous[0].Second.PreviousType.ToString().ToLower();
	                throw new ExprValidationException("The '" + funcname + "' function may not occur in the where-clause or having-clause of a statement with aggregations as 'previous' does not provide remove stream data; Use the 'first','last','window' or 'count' aggregation functions instead");
	            }
	        }

	        // Compile a map of aggregation nodes and equivalent-to aggregation nodes.
	        // Equivalent-to functions are for example "select sum(a*b), 5*sum(a*b)".
	        // Reducing the total number of aggregation functions.
	        IList<AggregationServiceAggExpressionDesc> aggregations = new List<AggregationServiceAggExpressionDesc>();
	        foreach (var selectAggNode in selectAggregateExprNodes)
	        {
	            AddEquivalent(selectAggNode, aggregations);
	        }
	        foreach (var havingAggNode in havingAggregateExprNodes)
	        {
	            AddEquivalent(havingAggNode, aggregations);
	        }
	        foreach (var orderByAggNode in orderByAggregateExprNodes)
	        {
	            AddEquivalent(orderByAggNode, aggregations);
	        }

	        // Construct a list of evaluation node for the aggregation functions (regular agg).
	        // For example "sum(2 * 3)" would make the sum an evaluation node.
	        IList<ExprEvaluator> methodAggEvaluatorsList = new List<ExprEvaluator>();
	        foreach (var aggregation in aggregations) {
	            var aggregateNode = aggregation.AggregationNode;
	            if (!aggregateNode.Factory.IsAccessAggregation) {
	                ExprEvaluator evaluator;
	                if (aggregateNode.ChildNodes.Length > 1) {
	                    evaluator = GetMultiNodeEvaluator(aggregateNode.ChildNodes, isJoin, typesPerStream);
	                }
	                else if (aggregateNode.ChildNodes.Length > 0) {
	                    if (aggregateNode.ChildNodes[0] is ExprWildcard) {
	                        evaluator = GetWildcardEvaluator(typesPerStream, isJoin);
	                    }
	                    else {
	                        // Use the evaluation node under the aggregation node to obtain the aggregation value
	                        evaluator = aggregateNode.ChildNodes[0].ExprEvaluator;
	                    }
	                }
	                // For aggregation that doesn't evaluate any particular sub-expression, return null on evaluation
	                else {
	                    evaluator = new ProxyExprEvaluator
                        {
	                        ProcEvaluate = args => null,
	                        ProcReturnType = () => null
	                    };
	                }
	                methodAggEvaluatorsList.Add(evaluator);
	            }
	        }

	        // determine binding
	        if (intoTableSpec != null) {

	            if (groupByRollupDesc != null) {
	                throw new ExprValidationException("Roll-up is not supported with tables");
	            }

	            // obtain metadata
	            var metadata = tableService.GetTableMetadata(intoTableSpec.Name);
	            if (metadata == null) {
	                throw new ExprValidationException("Invalid into-table clause: Failed to find table by name '" + intoTableSpec.Name + "'");
	            }

	            EPLValidationUtil.ValidateContextName(true, intoTableSpec.Name, metadata.ContextName, optionalContextName, false);

	            // validate group keys
	            var groupByTypes = ExprNodeUtility.GetExprResultTypes(groupByNodes);
	            ExprTableNodeUtil.ValidateExpressions(intoTableSpec.Name, groupByTypes, "group-by", groupByNodes,
	                    metadata.KeyTypes, "group-by");

	            // determine how this binds to existing aggregations, assign column numbers
	            BindingMatchResult bindingMatchResult = MatchBindingsAssignColumnNumbers(intoTableSpec, metadata, aggregations, selectClauseNamedNodes, methodAggEvaluatorsList, declaredExpressions);

	            // return factory
	            AggregationServiceFactory aggregationServiceFactory;
	            if (!hasGroupByClause) {
	                aggregationServiceFactory = factoryService.GetNoGroupWBinding(bindingMatchResult.Accessors, isJoin, bindingMatchResult.MethodPairs, intoTableSpec.Name, bindingMatchResult.TargetStates, bindingMatchResult.AccessStateExpr, bindingMatchResult.Agents);
	            }
	            else {
	                aggregationServiceFactory = factoryService.GetGroupWBinding(metadata, bindingMatchResult.MethodPairs, bindingMatchResult.Accessors, isJoin, intoTableSpec, bindingMatchResult.TargetStates, bindingMatchResult.AccessStateExpr, bindingMatchResult.Agents);
	            }
	            return new AggregationServiceFactoryDesc(aggregationServiceFactory, aggregations, groupKeyExpressions);
	        }

	        // Assign a column number to each aggregation node. The regular aggregation goes first followed by access-aggregation.
	        var columnNumber = 0;
	        foreach (var entry in aggregations)
	        {
	            if (!entry.Factory.IsAccessAggregation) {
	                entry.ColumnNum = columnNumber++;
	            }
	        }
	        foreach (var entry in aggregations)
	        {
	            if (entry.Factory.IsAccessAggregation) {
	                entry.ColumnNum = columnNumber++;
	            }
	        }

	        // determine method aggregation factories and evaluators(non-access)
	        ExprEvaluator[] methodAggEvaluators = methodAggEvaluatorsList.ToArray();
	        var methodAggFactories = new AggregationMethodFactory[methodAggEvaluators.Length];
	        var count = 0;
	        foreach (var aggregation in aggregations) {
	            var aggregateNode = aggregation.AggregationNode;
	            if (!aggregateNode.Factory.IsAccessAggregation) {
	                methodAggFactories[count] = aggregateNode.Factory;
	                count++;
	            }
	        }

	        // handle access aggregations
	        var multiFunctionAggPlan = AggregationMultiFunctionAnalysisHelper.AnalyzeAccessAggregations(aggregations);
	        var accessorPairs = multiFunctionAggPlan.AccessorPairs;
	        var accessAggregations = multiFunctionAggPlan.StateFactories;

	        AggregationServiceFactory serviceFactory;

	        // Handle without a group-by clause: we group all into the same pot
	        if (!hasGroupByClause) {
	            if ((methodAggEvaluators.Length > 0) && (accessorPairs.Length == 0)) {
	                serviceFactory = factoryService.GetNoGroupNoAccess(methodAggEvaluators, methodAggFactories);
	            }
	            else if ((methodAggEvaluators.Length == 0) && (accessorPairs.Length > 0)) {
	                serviceFactory = factoryService.GetNoGroupAccessOnly(accessorPairs, accessAggregations, isJoin);
	            }
	            else {
	                serviceFactory = factoryService.GetNoGroupAccessMixed(methodAggEvaluators, methodAggFactories, accessorPairs, accessAggregations, isJoin);
	            }
	        }
	        else {
	            var groupKeyBinding = methodResolutionService.GetGroupKeyBinding(groupByNodes, groupByRollupDesc);

	            var hasNoReclaim = HintEnum.DISABLE_RECLAIM_GROUP.GetHint(annotations) != null;
	            var reclaimGroupAged = HintEnum.RECLAIM_GROUP_AGED.GetHint(annotations);
	            var reclaimGroupFrequency = HintEnum.RECLAIM_GROUP_AGED.GetHint(annotations);
	            if (!isDisallowNoReclaim && hasNoReclaim)
	            {
	                if (groupByRollupDesc != null) {
	                    throw GetRollupReclaimEx();
	                }
	                if ((methodAggEvaluators.Length > 0) && (accessorPairs.Length == 0)) {
	                    serviceFactory = factoryService.GetGroupedNoReclaimNoAccess(methodAggEvaluators, methodAggFactories, groupKeyBinding);
	                }
	                else if ((methodAggEvaluators.Length == 0) && (accessorPairs.Length > 0)) {
	                    serviceFactory = factoryService.GetGroupNoReclaimAccessOnly(accessorPairs, accessAggregations, groupKeyBinding, isJoin);
	                }
	                else {
	                    serviceFactory = factoryService.GetGroupNoReclaimMixed(methodAggEvaluators, methodAggFactories, accessorPairs, accessAggregations, isJoin, groupKeyBinding);
	                }
	            }
	            else if (!isDisallowNoReclaim && reclaimGroupAged != null)
	            {
	                if (groupByRollupDesc != null) {
	                    throw GetRollupReclaimEx();
	                }
	                serviceFactory = factoryService.GetGroupReclaimAged(methodAggEvaluators, methodAggFactories, reclaimGroupAged, reclaimGroupFrequency, variableService, accessorPairs, accessAggregations, isJoin, groupKeyBinding, optionalContextName);
	            }
	            else if (groupByRollupDesc != null) {
	                serviceFactory = factoryService.GetGroupReclaimMixableRollup(methodAggEvaluators, methodAggFactories, accessorPairs, accessAggregations, isJoin, groupKeyBinding, groupByRollupDesc);
	            }
	            else
	            {
	                if ((methodAggEvaluators.Length > 0) && (accessorPairs.Length == 0)) {
	                    serviceFactory = factoryService.GetGroupReclaimNoAccess(methodAggEvaluators, methodAggFactories, accessorPairs, accessAggregations, isJoin, groupKeyBinding);
	                }
	                else {
	                    serviceFactory = factoryService.GetGroupReclaimMixable(methodAggEvaluators, methodAggFactories, accessorPairs, accessAggregations, isJoin, groupKeyBinding);
	                }
	            }
	        }

	        return new AggregationServiceFactoryDesc(serviceFactory, aggregations, groupKeyExpressions);
	    }

	    private static ExprEvaluator GetWildcardEvaluator(EventType[] typesPerStream, bool isJoin)
        {
	        var returnType = typesPerStream != null && typesPerStream.Length > 0 ? typesPerStream[0].UnderlyingType : null;
	        if (isJoin || returnType == null) {
	            throw new ExprValidationException("Invalid use of wildcard (*) for stream selection in a join or an empty from-clause, please use the stream-alias syntax to select a specific stream instead");
	        }
	        return new ProxyExprEvaluator() {
	            ProcEvaluate = args =>  {
	                EventBean @event = args.EventsPerStream[0];
	                if (@event == null) {
	                    return null;
	                }
	                return @event.Underlying;
	            },
	            ProcReturnType = () =>  {
	                return returnType;
	            },
	        };
	    }

	    private static BindingMatchResult MatchBindingsAssignColumnNumbers(
	        IntoTableSpec bindings,
	        TableMetadata metadata,
	        IList<AggregationServiceAggExpressionDesc> aggregations,
	        IDictionary<ExprNode, string> selectClauseNamedNodes,
	        IList<ExprEvaluator> methodAggEvaluatorsList,
	        IList<ExprDeclaredNode> declaredExpressions)
	    {
	        IDictionary<AggregationServiceAggExpressionDesc, TableMetadataColumnAggregation> methodAggs = new LinkedHashMap<AggregationServiceAggExpressionDesc, TableMetadataColumnAggregation>();
	        IDictionary<AggregationServiceAggExpressionDesc, TableMetadataColumnAggregation> accessAggs = new LinkedHashMap<AggregationServiceAggExpressionDesc, TableMetadataColumnAggregation>();
	        foreach (var aggDesc in aggregations) {

	            // determine assigned name
	            var columnName = FindColumnNameForAggregation(selectClauseNamedNodes, declaredExpressions, aggDesc.AggregationNode);
	            if (columnName == null) {
	                throw new ExprValidationException("Failed to find an expression among the select-clause expressions for expression '" + aggDesc.AggregationNode.ToExpressionStringMinPrecedence() + "'");
	            }

	            // determine binding metadata
	            var columnMetadata = (TableMetadataColumnAggregation) metadata.TableColumns.Get(columnName);
	            if (columnMetadata == null) {
	                throw new ExprValidationException("Failed to find name '" + columnName + "' among the columns for table '" + bindings.Name + "'");
	            }

	            // validate compatible
	            ValidateIntoTableCompatible(bindings.Name, columnName, columnMetadata, aggDesc);

	            if (!columnMetadata.Factory.IsAccessAggregation) {
	                methodAggs.Put(aggDesc, columnMetadata);
	            }
	            else {
	                accessAggs.Put(aggDesc, columnMetadata);
	            }
	        }

	        // handle method-aggs
	        var methodPairs = new TableColumnMethodPair[methodAggEvaluatorsList.Count];
	        var methodIndex = -1;
	        foreach (var methodEntry in methodAggs) {
	            methodIndex++;
	            var targetIndex = methodEntry.Value.MethodOffset;
	            methodPairs[methodIndex] = new TableColumnMethodPair(methodAggEvaluatorsList[methodIndex], targetIndex, methodEntry.Key.AggregationNode);
	            methodEntry.Key.ColumnNum = targetIndex;
	        }

	        // handle access-aggs
	        var accessSlots = new LinkedHashMap<int, ExprNode>();
	        var accessReadPairs = new List<AggregationAccessorSlotPair>();
	        var accessIndex = -1;
	        var agents = new List<AggregationAgent>();
	        foreach (var accessEntry in accessAggs) {
	            accessIndex++;
	            var slot = accessEntry.Value.AccessAccessorSlotPair.Slot;
	            var aggregationMethodFactory = accessEntry.Key.Factory;
	            var accessor = aggregationMethodFactory.Accessor;
	            accessSlots.Put(slot, accessEntry.Key.AggregationNode);
	            accessReadPairs.Add(new AggregationAccessorSlotPair(slot, accessor));
	            accessEntry.Key.ColumnNum = metadata.NumberMethodAggregations + accessIndex;
	            agents.Add(aggregationMethodFactory.AggregationStateAgent);
	        }
	        AggregationAgent[] agentArr= agents.ToArray();
	        AggregationAccessorSlotPair[] accessReads = accessReadPairs.ToArray();

	        var targetStates = new int[accessSlots.Count];
	        var accessStateExpr = new ExprNode[accessSlots.Count];
	        var count = 0;
	        foreach (var entry in accessSlots) {
	            targetStates[count] = entry.Key;
	            accessStateExpr[count] = entry.Value;
	            count++;
	        }

	        return new BindingMatchResult(methodPairs, accessReads, targetStates, accessStateExpr, agentArr);
	    }

	    private static string FindColumnNameForAggregation(
	        IDictionary<ExprNode, string> selectClauseNamedNodes,
	        IList<ExprDeclaredNode> declaredExpressions,
	        ExprAggregateNode aggregationNode)
        {
	        if (selectClauseNamedNodes.ContainsKey(aggregationNode)) {
	            return selectClauseNamedNodes.Get(aggregationNode);
	        }

	        foreach (var node in declaredExpressions) {
	            if (node.Body == aggregationNode) {
	                return node.Prototype.Name;
	            }
	        }

	        return null;
	    }

	    private static void ValidateIntoTableCompatible(
	        string tableName,
	        string columnName,
	        TableMetadataColumnAggregation columnMetadata,
	        AggregationServiceAggExpressionDesc aggDesc)
	    {
	        var factoryProvided = aggDesc.Factory;
	        var factoryRequired = columnMetadata.Factory;

	        try {
	            factoryRequired.ValidateIntoTableCompatible(factoryProvided);
	        }
	        catch (ExprValidationException ex) {
	            var text = GetMessage(tableName, columnName, factoryRequired.AggregationExpression, factoryProvided.AggregationExpression);
	            throw new ExprValidationException(text + ": " + ex.Message, ex);
	        }
	    }

	    private static string GetMessage(
	        string tableName,
	        string columnName,
	        ExprAggregateNodeBase aggregationRequired,
	        ExprAggregateNodeBase aggregationProvided)
	    {
	        return "Incompatible aggregation function for table '" +
	            tableName +
	            "' column '" +
	            columnName + "', expecting '" +
	            aggregationRequired.ToExpressionStringMinPrecedence() +
	            "' and received '" +
	            aggregationProvided.ToExpressionStringMinPrecedence() +
	            "'";
	    }

	    private static ExprEvaluator GetMultiNodeEvaluator(ExprNode[] childNodes, bool join, EventType[] typesPerStream)
	    {
	        var evaluators = new ExprEvaluator[childNodes.Length];

	        // determine constant nodes
	        var count = 0;
	        foreach (var node in childNodes) {
	            if (node is ExprWildcard) {
	                evaluators[count] = GetWildcardEvaluator(typesPerStream, join);
	            }
	            else {
	                evaluators[count] = node.ExprEvaluator;
	            }
	            count++;
	        }

	        return new ProxyExprEvaluator() {
	            ProcEvaluate = args =>
	            {
	                var values = new object[evaluators.Length];
	                for (var i = 0; i < evaluators.Length; i++) {
	                    values[i] = evaluators[i].Evaluate(args);
	                }
	                return values;
	            },

	            ProcReturnType = () =>  {
	                return typeof(object[]);
	            },
	        };
	    }

	    private static void AddEquivalent(ExprAggregateNode aggNodeToAdd, IList<AggregationServiceAggExpressionDesc> equivalencyList)
	    {
	        // Check any same aggregation nodes among all aggregation clauses
	        var foundEquivalent = false;
	        foreach (var existing in equivalencyList)
	        {
	            var aggNode = existing.AggregationNode;
	            if (ExprNodeUtility.DeepEquals(aggNode, aggNodeToAdd))
	            {
	                existing.AddEquivalent(aggNodeToAdd);
	                foundEquivalent = true;
	                break;
	            }
	        }

	        if (!foundEquivalent)
	        {
	            equivalencyList.Add(new AggregationServiceAggExpressionDesc(aggNodeToAdd, aggNodeToAdd.Factory));
	        }
	    }

	    public static ExprValidationException GetRollupReclaimEx()
        {
	        return new ExprValidationException("Reclaim hints are not available with rollup");
	    }

	    internal class BindingMatchResult
        {
	        internal BindingMatchResult(TableColumnMethodPair[] methodPairs, AggregationAccessorSlotPair[] accessors, int[] targetStates, ExprNode[] accessStateExpr, AggregationAgent[] agents) 
            {
	            MethodPairs = methodPairs;
	            Accessors = accessors;
	            TargetStates = targetStates;
	            AccessStateExpr = accessStateExpr;
	            Agents = agents;
	        }

	        public TableColumnMethodPair[] MethodPairs { get; private set; }

	        public AggregationAccessorSlotPair[] Accessors { get; private set; }

	        public int[] TargetStates { get; private set; }

	        public AggregationAgent[] Agents { get; private set; }

	        public ExprNode[] AccessStateExpr { get; private set; }
        }
	}
} // end of namespace
