///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2014 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Reflection;

using com.espertech.esper.epl.expression.core;

using XLR8.CGLib;

using com.espertech.esper.client;
using com.espertech.esper.compat;
using com.espertech.esper.compat.collections;
using com.espertech.esper.core.context.util;
using com.espertech.esper.epl.db;
using com.espertech.esper.epl.expression;
using com.espertech.esper.epl.spec;
using com.espertech.esper.events;
using com.espertech.esper.schedule;
using com.espertech.esper.util;
using com.espertech.esper.view;

namespace com.espertech.esper.epl.core
{
    using DataMap = IDictionary<string, object>;
    using TypeMap = IDictionary<string, object>;

    /// <summary>
    /// Factory for method-invocation data provider streams.
    /// </summary>
    public class MethodPollingViewableFactory
    {
        /// <summary>
        /// Creates a method-invocation polling view for use as a stream that calls a method, or pulls results from cache.
        /// </summary>
        /// <param name="streamNumber">the stream number</param>
        /// <param name="methodStreamSpec">defines the class and method to call</param>
        /// <param name="eventAdapterService">for creating event types and events</param>
        /// <param name="epStatementAgentInstanceHandle">for time-based callbacks</param>
        /// <param name="methodResolutionService">for resolving classes and imports</param>
        /// <param name="engineImportService">for resolving configurations</param>
        /// <param name="schedulingService">for scheduling callbacks in expiry-time based caches</param>
        /// <param name="scheduleBucket">for schedules within the statement</param>
        /// <param name="exprEvaluatorContext">expression evaluation context</param>
        /// <returns>pollable view</returns>
        /// <throws>ExprValidationException if the expressions cannot be validated or the method descriptorhas incorrect class and method names, or parameter number and types don't match
        /// </throws>
        public static HistoricalEventViewable CreatePollMethodView(
            int streamNumber,
            MethodStreamSpec methodStreamSpec,
            EventAdapterService eventAdapterService,
            EPStatementAgentInstanceHandle epStatementAgentInstanceHandle,
            MethodResolutionService methodResolutionService,
            EngineImportService engineImportService,
            SchedulingService schedulingService,
            ScheduleBucket scheduleBucket,
            ExprEvaluatorContext exprEvaluatorContext)
        {
            // Try to resolve the method
            FastMethod staticMethod;
            Type declaringClass;
            try
    		{
    			var method = methodResolutionService.ResolveMethod(methodStreamSpec.ClassName, methodStreamSpec.MethodName);
                declaringClass = method.DeclaringType;
                var declaringFastClass = FastClass.Create(method.DeclaringType);
    			staticMethod = declaringFastClass.GetMethod(method);
    		}
    		catch(Exception e)
    		{
    			throw new ExprValidationException(e.Message);
    		}
    
            // Determine object type returned by method
            var beanClass = staticMethod.ReturnType;
            if ((beanClass == typeof(void)) || (beanClass.IsBuiltinDataType()))
            {
                throw new ExprValidationException("Invalid return type for static method '" + staticMethod.Name + "' of class '" + methodStreamSpec.ClassName + "', expecting a class");
            }
            if (staticMethod.ReturnType.IsArray)
            {
                beanClass = staticMethod.ReturnType.GetElementType();
            }
    
            // If the method returns a Map, look up the map type
            IDictionary<String, Object> mapType = null;
            String mapTypeName = null;

            if ((staticMethod.ReturnType.IsGenericStringDictionary()) ||
                (staticMethod.ReturnType.IsArray && staticMethod.ReturnType.GetElementType().IsGenericStringDictionary()))
            {
                var metadata = GetCheckMetadata(methodStreamSpec.MethodName, methodStreamSpec.ClassName, methodResolutionService, typeof(IDictionary<string, object>));
                mapTypeName = metadata.TypeName;
                mapType = (IDictionary<String, Object>) metadata.TypeMetadata;
            }
    
            // If the method returns an Object[] or Object[][], look up the type information
            IDictionary<String, Object> oaType = null;
            String oaTypeName = null;
            if (staticMethod.ReturnType == typeof(Object[]) || staticMethod.ReturnType == typeof(Object[][]))
            {
                var metadata = GetCheckMetadata(methodStreamSpec.MethodName, methodStreamSpec.ClassName, methodResolutionService, typeof(IDictionary<string, object>));
                oaTypeName = metadata.TypeName;
                oaType = (IDictionary<String, Object>)metadata.TypeMetadata;
            }
    
            // Determine event type from class and method name
            EventType eventType;
            if (mapType != null) {
                eventType = eventAdapterService.AddNestableMapType(mapTypeName, mapType, null, false, true, true, false, false);
            }
            else if (oaType != null) {
                eventType = eventAdapterService.AddNestableObjectArrayType(oaTypeName, oaType, null, false, true, true, false, false, false, null);
            }
            else {
                eventType = eventAdapterService.AddBeanType(beanClass.FullName, beanClass, false, true, true);
            }
    
            // Construct polling strategy as a method invocation
            var configCache = engineImportService.GetConfigurationMethodRef(declaringClass.FullName);
            if (configCache == null)
            {
                configCache = engineImportService.GetConfigurationMethodRef(declaringClass.FullName);
            }
            var dataCacheDesc = (configCache != null) ? configCache.DataCacheDesc : null;
            var dataCache = DataCacheFactory.GetDataCache(dataCacheDesc, epStatementAgentInstanceHandle, schedulingService, scheduleBucket);
            PollExecStrategy methodPollStrategy;
            if (mapType != null) {
                if (staticMethod.ReturnType.IsArray()) {
                    methodPollStrategy = new MethodPollingExecStrategyMapArray(eventAdapterService, staticMethod, eventType);
                }
                else {
                    methodPollStrategy = new MethodPollingExecStrategyMapPlain(eventAdapterService, staticMethod, eventType);
                }
            }
            else if (oaType != null) {
                if (staticMethod.ReturnType == typeof(Object[][])) {
                    methodPollStrategy = new MethodPollingExecStrategyOAArray(eventAdapterService, staticMethod, eventType);
                }
                else {
                    methodPollStrategy = new MethodPollingExecStrategyOAPlain(eventAdapterService, staticMethod, eventType);
                }
            }
            else {
                if (staticMethod.ReturnType.IsArray()) {
                    methodPollStrategy = new MethodPollingExecStrategyPONOArray(eventAdapterService, staticMethod, eventType);
                }
                else {
                    methodPollStrategy = new MethodPollingExecStrategyPONOPlain(eventAdapterService, staticMethod, eventType);
                }
            }
    
            return new MethodPollingViewable(methodStreamSpec, streamNumber, methodStreamSpec.Expressions, methodPollStrategy, dataCache, eventType, exprEvaluatorContext);
        }
    
        private static MethodMetadataDesc GetCheckMetadata(String methodName, String className, MethodResolutionService methodResolutionService, Type metadataClass)
        {
            MethodInfo typeGetterMethod;
            var getterMethodName = methodName + "Metadata";
            try {
                typeGetterMethod = methodResolutionService.ResolveMethod(className, getterMethodName, new Type[0], new bool[0], new bool[0]);
            }
            catch(Exception) {
                throw new EPException("Could not find getter method for method invocation, expected a method by name '" + getterMethodName + "' accepting no parameters");
            }
    
            var fail = false;
            if (metadataClass.IsInterface) {
                fail = !typeGetterMethod.ReturnType.IsImplementsInterface(metadataClass);
            }
            else {
                fail = typeGetterMethod.ReturnType != metadataClass;
            }
            if (fail) {
                throw new EPException("Getter method '" + getterMethodName + "' does not return " + metadataClass.GetTypeNameFullyQualPretty());
            }
    
            Object resultType;
            try
            {
                resultType = typeGetterMethod.Invoke(null, null);
            }
            catch (Exception e) {
                throw new EPException("Error invoking metadata getter method for method invocation, for method by name '" + getterMethodName + "' accepting no parameters: " + e.Message, e);
            }
            if (resultType == null) {
                throw new EPException("Error invoking metadata getter method for method invocation, method returned a null value");
            }
    
            return new MethodMetadataDesc(className + "." + typeGetterMethod.Name, resultType);
        }
    
        public struct MethodMetadataDesc
        {
            public MethodMetadataDesc(String typeName, Object typeMetadata)
            {
                TypeName = typeName;
                TypeMetadata = typeMetadata;
            }

            public string TypeName;
            public object TypeMetadata;
        }
    }
}
