///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2014 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;

using com.espertech.esper.client;
using com.espertech.esper.compat;
using com.espertech.esper.core.context.util;
using com.espertech.esper.epl.agg.rollup;
using com.espertech.esper.epl.expression.core;
using com.espertech.esper.epl.expression;
using com.espertech.esper.epl.variable;

namespace com.espertech.esper.epl.core
{
    /// <summary>
    /// An order-by processor that sorts events according to the expressions
    /// in the order_by clause.
    /// </summary>
    public class OrderByProcessorRowLimit : OrderByProcessor
    {
        private readonly VariableReader _numRowsVariableReader;
        private readonly VariableReader _offsetVariableReader;
        private int _currentRowLimit;
        private int _currentOffset;
    
        public OrderByProcessorRowLimit(VariableReader numRowsVariableReader, VariableReader offsetVariableReader, int currentRowLimit, int currentOffset)
        {
            _numRowsVariableReader = numRowsVariableReader;
            _offsetVariableReader = offsetVariableReader;
            _currentRowLimit = currentRowLimit;
            _currentOffset = currentOffset;
        }
    
        public EventBean[] Sort(EventBean[] outgoingEvents, EventBean[][] generatingEvents, bool isNewData, ExprEvaluatorContext exprEvaluatorContext)
        {
            return ApplyLimit(outgoingEvents);
        }
    
        public EventBean[] Sort(EventBean[] outgoingEvents, EventBean[][] generatingEvents, Object[] groupByKeys, bool isNewData, ExprEvaluatorContext exprEvaluatorContext)
        {
            return ApplyLimit(outgoingEvents);
        }
    
        public EventBean[] Sort(EventBean[] outgoingEvents, IList<GroupByRollupKey> currentGenerators, bool newData, AgentInstanceContext agentInstanceContext, OrderByElement[][] elementsPerLevel) {
            return ApplyLimit(outgoingEvents);
        }
    
        public Object GetSortKey(EventBean[] eventsPerStream, bool isNewData, ExprEvaluatorContext exprEvaluatorContext)
        {
            return null;
        }
    
        public Object GetSortKey(EventBean[] eventsPerStream, bool isNewData, ExprEvaluatorContext exprEvaluatorContext, OrderByElement[] elementsForLevel) {
            return null;
        }
    
        public Object[] GetSortKeyPerRow(EventBean[] generatingEvents, bool isNewData, ExprEvaluatorContext exprEvaluatorContext)
        {
            return null;
        }
    
        public EventBean[] Sort(EventBean[] outgoingEvents, Object[] orderKeys, ExprEvaluatorContext exprEvaluatorContext)
        {
            return ApplyLimit(outgoingEvents);
        }
    
        /// <summary>
        /// Applys the limiting function to outgoing events.
        /// </summary>
        /// <param name="outgoingEvents">unlimited</param>
        /// <returns>limited</returns>
        protected internal EventBean[] ApplyLimit(EventBean[] outgoingEvents)
        {
            if (outgoingEvents == null) {
                return null;
            }
            if (_numRowsVariableReader != null)
            {
                var varValue = _numRowsVariableReader.Value;
                if (varValue != null)
                {
                    _currentRowLimit = varValue.AsInt();
                }
                else
                {
                    _currentRowLimit = int.MaxValue;
                }
                if (_currentRowLimit < 0)
                {
                    _currentRowLimit = int.MaxValue;
                }
            }
    
            if (_offsetVariableReader != null)
            {
                var varValue = _offsetVariableReader.Value;
                if (varValue != null)
                {
                    _currentOffset = varValue.AsInt();
                }
                else
                {
                    _currentOffset = 0;
                }
                if (_currentOffset < 0)
                {
                    _currentOffset = 0;
                }
            }
    
            // no offset
            if (_currentOffset == 0)
            {
                if ((outgoingEvents.Length <= _currentRowLimit))
                {
                    return outgoingEvents;
                }
    
                if (_currentRowLimit == 0)
                {
                    return null;
                }
    
                var limited = new EventBean[_currentRowLimit];
                Array.Copy(outgoingEvents, 0, limited, 0, _currentRowLimit);
                return limited;
            }
            // with offset
            else
            {
                var maxInterested = _currentRowLimit + _currentOffset;
                if (_currentRowLimit == int.MaxValue)
                {
                    maxInterested = int.MaxValue;
                }
    
                // more rows then requested
                if (outgoingEvents.Length > maxInterested)
                {
                    var limitedX = new EventBean[_currentRowLimit];
                    Array.Copy(outgoingEvents, _currentOffset, limitedX, 0, _currentRowLimit);
                    return limitedX;
                }
    
                // less or equal rows to offset
                if (outgoingEvents.Length <= _currentOffset)
                {
                    return null;
                }
    
                var size = outgoingEvents.Length - _currentOffset;
                var limited = new EventBean[size];
                Array.Copy(outgoingEvents, _currentOffset, limited, 0, size);
                return limited;
            }
        }
    }
}
