///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2014 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;

using com.espertech.esper.core.context.util;
using com.espertech.esper.epl.agg.service;
using com.espertech.esper.epl.expression;
using com.espertech.esper.epl.expression.core;
using com.espertech.esper.epl.spec;
using com.espertech.esper.epl.variable;
using com.espertech.esper.util;

namespace com.espertech.esper.epl.core
{
    /// <summary>
    /// An order-by processor that sorts events according to the expressions
    /// in the order_by clause.
    /// </summary>
    public class OrderByProcessorRowLimitFactory : OrderByProcessorFactory
    {
        private readonly VariableMetaData _numRowsVariableMetaData;
        private readonly VariableMetaData _offsetVariableMetaData;
        private readonly int _currentRowLimit;
        private readonly int _currentOffset;

        /// <summary>
        /// Ctor.
        /// </summary>
        /// <param name="rowLimitSpec">specification for row limit, or null if no row limit is defined</param>
        /// <param name="variableService">for retrieving variable state for use with row limiting</param>
        /// <param name="optionalContextName">Name of the optional context.</param>
        /// <throws><seealso cref="ExprValidationException" /> if row limit specification validation fails</throws>
        public OrderByProcessorRowLimitFactory(RowLimitSpec rowLimitSpec, VariableService variableService, String optionalContextName)
        {
            if (rowLimitSpec.NumRowsVariable != null)
            {
                _numRowsVariableMetaData = variableService.GetVariableMetaData(rowLimitSpec.NumRowsVariable);
                if (_numRowsVariableMetaData == null) {
                    throw new ExprValidationException("Limit clause variable by name '" + rowLimitSpec.NumRowsVariable + "' has not been declared");
                }
                var message = VariableServiceUtil.CheckVariableContextName(optionalContextName, _numRowsVariableMetaData);
                if (message != null) {
                    throw new ExprValidationException(message);
                }
                if (!_numRowsVariableMetaData.VariableType.IsNumeric())
                {
                    throw new ExprValidationException("Limit clause requires a variable of numeric type");
                }
            }
            else
            {
                _numRowsVariableMetaData = null;
                _currentRowLimit = rowLimitSpec.NumRows.GetValueOrDefault(int.MaxValue);
                if (_currentRowLimit < 0)
                {
                    _currentRowLimit = int.MaxValue;
                }
            }
    
            if (rowLimitSpec.OptionalOffsetVariable != null)
            {
                _offsetVariableMetaData = variableService.GetVariableMetaData(rowLimitSpec.OptionalOffsetVariable);
                if (_offsetVariableMetaData == null) {
                    throw new ExprValidationException("Limit clause variable by name '" + rowLimitSpec.OptionalOffsetVariable + "' has not been declared");
                }
                var message = VariableServiceUtil.CheckVariableContextName(optionalContextName, _offsetVariableMetaData);
                if (message != null) {
                    throw new ExprValidationException(message);
                }
                if (!_offsetVariableMetaData.VariableType.IsNumeric())
                {
                    throw new ExprValidationException("Limit clause requires a variable of numeric type");
                }
            }
            else
            {
                _offsetVariableMetaData = null;
                if (rowLimitSpec.OptionalOffset != null)
                {
                    _currentOffset = rowLimitSpec.OptionalOffset.GetValueOrDefault(0);
    
                    if (_currentOffset <= 0)
                    {
                        throw new ExprValidationException("Limit clause requires a positive offset");
                    }
                }
                else
                {
                    _currentOffset = 0;
                }
            }
        }
    
        public OrderByProcessor Instantiate(AggregationService aggregationService, AgentInstanceContext agentInstanceContext)
        {
            VariableReader numRowsVariableReader = null;
            if (_numRowsVariableMetaData != null) {
                numRowsVariableReader = agentInstanceContext.StatementContext.VariableService.GetReader(
                    _numRowsVariableMetaData.VariableName, agentInstanceContext.AgentInstanceId);
            }
    
            VariableReader offsetVariableReader = null;
            if (_offsetVariableMetaData != null) {
                offsetVariableReader = agentInstanceContext.StatementContext.VariableService.GetReader(
                    _offsetVariableMetaData.VariableName, agentInstanceContext.AgentInstanceId);
            }
    
            return new OrderByProcessorRowLimit(numRowsVariableReader, offsetVariableReader,
                    _currentRowLimit, _currentOffset);
        }
    }
}
