///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2012 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.IO;

using com.espertech.esper.compat;
using com.espertech.esper.epl.expression.core;
using com.espertech.esper.metrics.instrumentation;
using com.espertech.esper.util;

namespace com.espertech.esper.epl.expression.funcs
{
    /// <summary>
    /// Represents the CAST(expression, type) function is an expression tree.
    /// </summary>
    [Serializable]
    public class ExprCastNode : ExprNodeBase, ExprEvaluator
    {
        private readonly String _typeIdentifier;
        private TypeCaster _typeCaster;
        private Type _targetType;
        private Type _sourceType;

        [NonSerialized] private ExprEvaluator _evaluator;

        /// <summary>Ctor. </summary>
        /// <param name="typeIdentifier">the the name of the type to cast to</param>
        public ExprCastNode(String typeIdentifier)
        {
            _typeIdentifier = typeIdentifier;
        }

        public override ExprEvaluator ExprEvaluator
        {
            get { return this; }
        }

        /// <summary>Returns the name of the type of cast to. </summary>
        /// <value>type name</value>
        public string TypeIdentifier
        {
            get { return _typeIdentifier; }
        }

        public override ExprNode Validate(ExprValidationContext validationContext)
        {
            if (ChildNodes.Length != 1)
            {
                throw new ExprValidationException("Cast function node must have exactly 1 child node");
            }
    
            _evaluator = ChildNodes[0].ExprEvaluator;
            _sourceType = _evaluator.ReturnType;

            // try the primitive names including "string"
            _targetType = TypeHelper.GetPrimitiveTypeForName(_typeIdentifier);
            if (_targetType == null)
            {
                try
                {
                    _targetType = TypeHelper.ResolveType(_typeIdentifier, true);
                }
                catch (Exception e)
                {
                    throw new ExprValidationException(
                        "Type as listed in cast function by name '" + _typeIdentifier + "' cannot be loaded", e);
                }
            }

            _sourceType = _sourceType.GetBoxedType();
            _targetType = _targetType.GetBoxedType();

            if (_sourceType == _targetType)
            {
                _typeCaster = o => o;
            }
            else if (_targetType == typeof(string))
            {
                _typeCaster = o => Convert.ToString(o);
            }
            else if (_sourceType == typeof(string))
            {
                var typeParser = SimpleTypeParserFactory.GetParser(_targetType);
                _typeCaster = o => typeParser((string)o);
            }
            else
            {
                _typeCaster = CastHelper.GetTypeCaster(_sourceType, _targetType);
            }

            return null;
        }

        public override bool IsConstantResult
        {
            get { return false; }
        }

        public Type ReturnType
        {
            get { return _targetType; }
        }

        public Object Evaluate(EvaluateParams evaluateParams)
        {
            Mutable<object> result = new Mutable<object>();

            using (Instrument.With(
                i => i.QExprCast(this),
                i => i.AExprCast(result)))
            {
                result.Value = _evaluator.Evaluate(evaluateParams);
                if (result.Value == null)
                {
                    return null;
                }

                if (_typeCaster != null)
                {
                    result.Value = _typeCaster.Invoke(result.Value);
                    if (result.Value != null)
                    {
                        return result.Value;
                    }
                }

                // Simple dynamic-cast
                if (_targetType.IsInstanceOfType(result.Value))
                {
                    if (!_targetType.IsValueType)
                    {
                        return result.Value;
                    }
                }

                try
                {
                    result.Value = Convert.ChangeType(result.Value, _targetType);
                    return result.Value;
                }
                catch (FormatException) // Occurs when numeric types parse strings
                {
                    result.Value = null;
                    return null;
                }
                catch (InvalidCastException)
                {
                    result.Value = null;
                    return null;
                }
            }
        }

        public override void ToPrecedenceFreeEPL(TextWriter writer)
        {
            writer.Write("cast(");
            ChildNodes[0].ToEPL(writer, ExprPrecedenceEnum.MINIMUM);
            writer.Write(",");
            writer.Write(_typeIdentifier);
            writer.Write(')');
        }

        public override ExprPrecedenceEnum Precedence
        {
            get { return ExprPrecedenceEnum.UNARY; }
        }

        public override bool EqualsNode(ExprNode node)
        {
            var other = node as ExprCastNode;
            if (other == null)
            {
                return false;
            }

            return _typeIdentifier == other._typeIdentifier;
        }
    }
}
