///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2012 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.Linq;

using com.espertech.esper.client;
using com.espertech.esper.compat;
using com.espertech.esper.compat.collections;
using com.espertech.esper.epl.join.rep;
using com.espertech.esper.util;

namespace com.espertech.esper.epl.join.assemble
{
    /// <summary>
    /// Represents a node in a tree responsible for assembling outer join query results.
    /// 
    /// The tree is double-linked, child nodes know each parent and parent know all child nodes.
    /// 
    /// Each specific subclass of this abstract assembly node is dedicated to assembling results for
    /// a certain event stream.
    /// </summary>

    public abstract class BaseAssemblyNode : ResultAssembler
    {
        /// <summary> Returns the stream number.</summary>
        /// <returns> stream number
        /// </returns>

        virtual public int StreamNum
        {
            get { return _streamNum; }
        }

        /// <summary> Returns an array of stream numbers that lists all child node's stream numbers.</summary>
        /// <returns> child node stream numbers
        /// </returns>
        virtual public int[] Substreams
        {
            get
            {
                IList<Int32> substreams = new List<Int32>();
                RecusiveAddSubstreams(substreams);

                // copy to array
                int[] substreamArr = new int[substreams.Count];
                int count = 0;

                foreach (Int32 stream in substreams)
                {
                    substreamArr[count++] = stream;
                }

                return substreamArr;
            }

        }
        /// <summary> Parent node.</summary>
        protected ResultAssembler ParentNode;

        /// <summary> Child nodes.</summary>
        private readonly IList<BaseAssemblyNode> _childNodes;

        /// <summary> Stream number.</summary>
        private readonly int _streamNum;

        /// <summary> Number of streams in statement.</summary>
        private readonly int _numStreams;

        /// <summary> Ctor.</summary>
        /// <param name="streamNum">stream number of the event stream that this node assembles results for.
        /// </param>
        /// <param name="numStreams">number of streams
        /// </param>

        protected BaseAssemblyNode(int streamNum, int numStreams)
        {
            _streamNum = streamNum;
            _numStreams = numStreams;
            _childNodes = new List<BaseAssemblyNode>();
        }

        /// <summary> Provides results to assembly nodes for initialization.</summary>
        /// <param name="result">is a list of result nodes per stream
        /// </param>
        public abstract void Init(IList<Node>[] result);

        /// <summary>
        /// Process results.
        /// </summary>
        /// <param name="result">is a list of result nodes per stream</param>
        /// <param name="resultFinalRows">The result final rows.</param>
        /// <param name="resultRootEvent">The result root event.</param>
        public abstract void Process(IList<Node>[] result, ICollection<EventBean[]> resultFinalRows, EventBean resultRootEvent);

        /// <summary> Output this node using writer, not outputting child nodes.</summary>
        /// <param name="indentWriter">to use for output
        /// </param>
        public abstract void Print(IndentWriter indentWriter);

        /// <summary> Add a child node.</summary>
        /// <param name="childNode">to add
        /// </param>
        public virtual void AddChild(BaseAssemblyNode childNode)
        {
            childNode.ParentAssembler = this;
            _childNodes.Add(childNode);
        }

        internal int NumStreams
        {
            get { return _numStreams; }
        }

        /// <summary> Returns child nodes.</summary>
        /// <returns> child nodes
        /// </returns>
        public IList<BaseAssemblyNode> ChildNodes
        {
            get
            {
                return _childNodes;
            }
        }

        /// <summary> Gets or sets the parent node.</summary>
        /// <returns> parent node
        /// </returns>
        public virtual ResultAssembler ParentAssembler
        {
            get { return ParentNode; }
            set { ParentNode = value; }
        }

        private void RecusiveAddSubstreams(IList<Int32> substreams)
        {
            substreams.Add(_streamNum);

            foreach (BaseAssemblyNode child in _childNodes)
            {
                child.RecusiveAddSubstreams(substreams);
            }
        }

        /// <summary> Output this node and all descendent nodes using writer, outputting child nodes.</summary>
        /// <param name="indentWriter">to output to
        /// </param>
        public virtual void PrintDescendends(IndentWriter indentWriter)
        {
            Print(indentWriter);
            foreach (BaseAssemblyNode child in _childNodes)
            {
                indentWriter.IncrIndent();
                child.Print(indentWriter);
                indentWriter.DecrIndent();
            }
        }

        /// <summary> Returns all descendent nodes to the top node in a list in which the utmost descendants are
        /// listed first and the top node itself is listed last.
        /// </summary>
        /// <param name="topNode">is the root node of a tree structure
        /// </param>
        /// <returns> list of nodes with utmost descendants first ordered by level of depth in tree with top node last
        /// </returns>
        public static IList<BaseAssemblyNode> GetDescendentNodesBottomUp(BaseAssemblyNode topNode)
        {
            var result = new List<BaseAssemblyNode>();

            // Map to hold per level of the node (1 to N depth) of node a list of nodes, if any
            // exist at that level
            var nodesPerLevel = new OrderedDictionary<Int32, IList<BaseAssemblyNode>>();

            // Recursively enter all aggregate functions and their level into map
            RecursiveAggregateEnter(topNode, nodesPerLevel, 1);

            // Done if none found
            if (nodesPerLevel.Count == 0)
            {
                throw new IllegalStateException("Empty collection for nodes per level");
            }

            // From the deepest (highest) level to the lowest, add aggregates to list
            int deepLevel = nodesPerLevel.Keys.Last();
            for (int i = deepLevel; i >= 1; i--)
            {
                IList<BaseAssemblyNode> list;
                if (!nodesPerLevel.TryGetValue(i, out list))
                {
                    continue;
                }

                result.AddRange(list);
            }

            return result;
        }

        private static void RecursiveAggregateEnter(BaseAssemblyNode currentNode, IDictionary<Int32, IList<BaseAssemblyNode>> nodesPerLevel, int currentLevel)
        {
            // ask all child nodes to enter themselves
            foreach (BaseAssemblyNode node in currentNode.ChildNodes)
            {
                RecursiveAggregateEnter(node, nodesPerLevel, currentLevel + 1);
            }

            // Add myself to list
            IList<BaseAssemblyNode> aggregates = nodesPerLevel.Get(currentLevel, null);
            if (aggregates == null)
            {
                aggregates = new List<BaseAssemblyNode>();
                nodesPerLevel[currentLevel] = aggregates;
            }
            aggregates.Add(currentNode);
        }

        public abstract void Result(EventBean[] row,
                                    int fromStreamNum,
                                    EventBean myEvent,
                                    Node myNode,
                                    ICollection<EventBean[]> resultFinalRows,
                                    EventBean resultRootEvent);
    }
}
