///////////////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2006-2014 Esper Team. All rights reserved.                           /
// http://esper.codehaus.org                                                          /
// ---------------------------------------------------------------------------------- /
// The software in this package is published under the terms of the GPL license       /
// a copy of which has been included with this distribution in the license.txt file.  /
///////////////////////////////////////////////////////////////////////////////////////

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

using com.espertech.esper.client;
using com.espertech.esper.collection;
using com.espertech.esper.compat;
using com.espertech.esper.compat.collections;
using com.espertech.esper.compat.logging;
using com.espertech.esper.core.context.util;
using com.espertech.esper.core.service;
using com.espertech.esper.core.start;
using com.espertech.esper.epl.core;
using com.espertech.esper.epl.expression.baseagg;
using com.espertech.esper.epl.expression.core;
using com.espertech.esper.epl.expression.funcs;
using com.espertech.esper.epl.expression.ops;
using com.espertech.esper.epl.expression.subquery;
using com.espertech.esper.epl.expression.visitor;
using com.espertech.esper.epl.named;
using com.espertech.esper.epl.property;
using com.espertech.esper.epl.spec;
using com.espertech.esper.epl.table.mgmt;
using com.espertech.esper.epl.variable;
using com.espertech.esper.events;
using com.espertech.esper.events.property;
using com.espertech.esper.schedule;
using com.espertech.esper.script;
using com.espertech.esper.type;
using com.espertech.esper.util;
using com.espertech.esper.view;

namespace com.espertech.esper.filter
{
	/// <summary>
	/// Helper to compile (validate and optimize) filter expressions as used in pattern and filter-based streams.
	/// </summary>
	public class FilterSpecCompiler
	{
        private static readonly ILog Log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

	    /// <summary>
	    /// Assigned for filter parameters that are based on boolean expression and not on
	    /// any particular property name.
	    /// <para />Keeping this artificial property name is a simplification as optimized filter parameters
	    /// generally keep a property name.
	    /// </summary>
	    public const string PROPERTY_NAME_BOOLEAN_EXPRESSION = ".boolean_expression";

        /// <summary>
        /// Factory method for compiling filter expressions into a filter specification
        /// for use with filter service.
        /// </summary>
        /// <param name="eventType">is the filtered-out event type</param>
        /// <param name="eventTypeName">is the name of the event type</param>
        /// <param name="filterExpessions">is a list of filter expressions</param>
        /// <param name="optionalPropertyEvalSpec">specification for evaluating properties</param>
        /// <param name="taggedEventTypes">is a map of stream names (tags) and event types available</param>
        /// <param name="arrayEventTypes">is a map of name tags and event type per tag for repeat-expressions that generate an array of events</param>
        /// <param name="streamTypeService">is used to set rules for resolving properties</param>
        /// <param name="optionalStreamName">the stream name, if provided</param>
        /// <param name="statementContext">context for statement</param>
        /// <param name="assignedTypeNumberStack">The assigned type number stack.</param>
        /// <returns>
        /// compiled filter specification
        /// </returns>
        /// <throws>ExprValidationException if the expression or type validations failed</throws>
	    public static FilterSpecCompiled MakeFilterSpec(
	        EventType eventType,
	        string eventTypeName,
	        IList<ExprNode> filterExpessions,
	        PropertyEvalSpec optionalPropertyEvalSpec,
            IDictionary<string, Pair<EventType, string>> taggedEventTypes,
            IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        StreamTypeService streamTypeService,
	        string optionalStreamName,
	        StatementContext statementContext,
	        ICollection<int> assignedTypeNumberStack)
	    {
	        // Validate all nodes, make sure each returns a boolean and types are good;
	        // Also decompose all AND super nodes into individual expressions
	        var validatedNodes = ValidateAllowSubquery(ExprNodeOrigin.FILTER, filterExpessions, streamTypeService, statementContext, taggedEventTypes, arrayEventTypes);
	        return Build(validatedNodes, eventType, eventTypeName, optionalPropertyEvalSpec, taggedEventTypes, arrayEventTypes, streamTypeService, optionalStreamName, statementContext, assignedTypeNumberStack);
	    }

	    public static FilterSpecCompiled Build(
	        IList<ExprNode> validatedNodes,
	        EventType eventType,
	        string eventTypeName,
	        PropertyEvalSpec optionalPropertyEvalSpec,
            IDictionary<string, Pair<EventType, string>> taggedEventTypes,
            IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        StreamTypeService streamTypeService,
	        string optionalStreamName,
	        StatementContext stmtContext,
	        ICollection<int> assignedTypeNumberStack)
        {
	        var evaluatorContextStmt = new ExprEvaluatorContextStatement(stmtContext, false);

	        return BuildNoStmtCtx(validatedNodes, eventType, eventTypeName, optionalPropertyEvalSpec, taggedEventTypes, arrayEventTypes, streamTypeService,
	                optionalStreamName, assignedTypeNumberStack,
	                evaluatorContextStmt, 
                    stmtContext.StatementId, 
                    stmtContext.StatementName, 
                    stmtContext.Annotations, 
                    stmtContext.ContextDescriptor,
	                stmtContext.MethodResolutionService, 
                    stmtContext.EventAdapterService, 
                    stmtContext.TimeProvider, 
                    stmtContext.VariableService,
                    stmtContext.ScriptingService,
                    stmtContext.TableService, 
                    stmtContext.ConfigSnapshot, 
                    stmtContext.NamedWindowService);
	    }

	    public static FilterSpecCompiled BuildNoStmtCtx(
	        IList<ExprNode> validatedNodes,
	        EventType eventType,
	        string eventTypeName,
	        PropertyEvalSpec optionalPropertyEvalSpec,
            IDictionary<string, Pair<EventType, string>> taggedEventTypes,
            IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        StreamTypeService streamTypeService,
	        string optionalStreamName,
	        ICollection<int> assignedTypeNumberStack,
	        ExprEvaluatorContext exprEvaluatorContext,
	        string statementId,
	        string statementName,
	        Attribute[] annotations,
	        ContextDescriptor contextDescriptor,
	        MethodResolutionService methodResolutionService,
	        EventAdapterService eventAdapterService,
	        TimeProvider timeProvider,
	        VariableService variableService,
            ScriptingService scriptingService,
	        TableService tableService,
	        ConfigurationInformation configurationInformation,
	        NamedWindowService namedWindowService)
        {
	        var constituents = DecomposeCheckAggregation(validatedNodes);

	        // From the constituents make a filter specification
	        var filterParamExprMap = new FilterParamExprMap();

	        // Make filter parameter for each expression node, if it can be optimized

	        foreach (var constituent in constituents)
	        {
	            var param = MakeFilterParam(constituent, taggedEventTypes, arrayEventTypes, exprEvaluatorContext, statementName);
	            filterParamExprMap.Put(constituent, param); // accepts null values as the expression may not be optimized
	        }

	        // Consolidate entries as possible, i.e. (a != 5 and a != 6) is (a not in (5,6))
	        // Removes duplicates for same property and same filter operator for filter service index optimizations
	        Consolidate(filterParamExprMap, statementName);

	        // Use all filter parameter and unassigned expressions
	        IList<FilterSpecParam> filterParams = new List<FilterSpecParam>();
	        filterParams.AddAll(filterParamExprMap.FilterParams);
	        var remainingExprNodes = filterParamExprMap.UnassignedExpressions;

	        // any unoptimized expression nodes are put under one AND
	        ExprNode exprNode = null;
	        if (!remainingExprNodes.IsEmpty())
	        {
	            if (remainingExprNodes.Count == 1)
	            {
	                exprNode = remainingExprNodes[0];
	            }
	            else
	            {
	                var andNode = ExprNodeUtility.ConnectExpressionsByLogicalAnd(remainingExprNodes);
	                var validationContext = new ExprValidationContext(streamTypeService, methodResolutionService, null, timeProvider, variableService, tableService, exprEvaluatorContext, eventAdapterService, statementName, statementId, annotations, contextDescriptor, scriptingService, false, false, true, false, null);
	                andNode.Validate(validationContext);
	                exprNode = andNode;
	            }
	        }

	        // if there are boolean expressions, add
	        if (exprNode != null)
	        {
	            var hasSubselectFilterStream = DetermineSubselectFilterStream(exprNode);
	            var hasTableAccess = DetermineTableAccessFilterStream(exprNode);
	            var lookupable = new FilterSpecLookupable(PROPERTY_NAME_BOOLEAN_EXPRESSION, null, exprNode.ExprEvaluator.ReturnType);
	            var param = new FilterSpecParamExprNode(lookupable, FilterOperator.BOOLEAN_EXPRESSION, exprNode, taggedEventTypes, arrayEventTypes, variableService, tableService, eventAdapterService, configurationInformation, statementName, hasSubselectFilterStream, hasTableAccess);
	            filterParams.Add(param);
	        }

	        PropertyEvaluator optionalPropertyEvaluator = null;
	        if (optionalPropertyEvalSpec != null)
	        {
	            optionalPropertyEvaluator = PropertyEvaluatorFactory.MakeEvaluator(
	                optionalPropertyEvalSpec, eventType, optionalStreamName, eventAdapterService, methodResolutionService,
	                timeProvider, variableService, scriptingService, tableService, streamTypeService.EngineURIQualifier, statementId,
	                statementName, annotations, assignedTypeNumberStack, configurationInformation, namedWindowService);
	        }

	        var spec = new FilterSpecCompiled(eventType, eventTypeName, filterParams, optionalPropertyEvaluator);

	        if (Log.IsDebugEnabled)
	        {
	            Log.Debug(".makeFilterSpec spec=" + spec);
	        }

	        return spec;
	    }

	    private static bool DetermineTableAccessFilterStream(ExprNode exprNode)
        {
	        var visitor = new ExprNodeTableAccessFinderVisitor();
	        exprNode.Accept(visitor);
	        return visitor.HasTableAccess;
	    }

	    private static bool DetermineSubselectFilterStream(ExprNode exprNode)
        {
	        var visitor = new ExprNodeSubselectDeclaredDotVisitor();
	        exprNode.Accept(visitor);
	        if (visitor.Subselects.IsEmpty()) {
	            return false;
	        }

	        return visitor.Subselects.Any(subselectNode => subselectNode.IsFilterStreamSubselect);
        }

	    // remove duplicate propertyName + filterOperator items making a judgement to optimize or simply remove the optimized form
	    private static void Consolidate(IList<FilterSpecParam> items, FilterParamExprMap filterParamExprMap, string statementName)
	    {
	        var op = items[0].FilterOperator;
	        if (op == FilterOperator.NOT_EQUAL)
	        {
	            HandleConsolidateNotEqual(items, filterParamExprMap, statementName);
	        }
	        else
	        {
	            // for all others we simple remove the second optimized form (filter param with same prop name and filter op)
	            // and thus the boolean expression that started this is included
	            for (var i = 1; i < items.Count; i++)
	            {
	                filterParamExprMap.RemoveValue(items[i]);
	            }
	        }
	    }

        /// <summary>
        /// Validates expression nodes and returns a list of validated nodes.
        /// </summary>
        /// <param name="exprNodeOrigin">The expr node origin.</param>
        /// <param name="exprNodes">is the nodes to validate</param>
        /// <param name="streamTypeService">is provding type information for each stream</param>
        /// <param name="statementContext">context</param>
        /// <param name="taggedEventTypes">pattern tagged types</param>
        /// <param name="arrayEventTypes">@return list of validated expression nodes</param>
        /// <returns>
        /// expr nodes
        /// </returns>
        /// <exception cref="ExprValidationException">
        /// Failed to validate  + EPStatementStartMethodHelperSubselect.GetSubqueryInfoText(count, subselect) + :  + ex.Message
        /// or
        /// Filter expression not returning a boolean value: ' + ExprNodeUtility.ToExpressionStringMinPrecedence(validated) + '
        /// </exception>
        /// <throws>ExprValidationException for validation errors</throws>
	    public static IList<ExprNode> ValidateAllowSubquery(
	        ExprNodeOrigin exprNodeOrigin,
	        IList<ExprNode> exprNodes,
	        StreamTypeService streamTypeService,
	        StatementContext statementContext,
	        IDictionary<string, Pair<EventType, string>> taggedEventTypes,
	        IDictionary<string, Pair<EventType, string>> arrayEventTypes)
	    {
	        IList<ExprNode> validatedNodes = new List<ExprNode>();

	        var evaluatorContextStmt = new ExprEvaluatorContextStatement(statementContext, false);
            var validationContext = new ExprValidationContext(
                streamTypeService, statementContext.MethodResolutionService, null, statementContext.TimeProvider,
                statementContext.VariableService, 
                statementContext.TableService, 
                evaluatorContextStmt,
                statementContext.EventAdapterService,
                statementContext.StatementName, 
                statementContext.StatementId,
                statementContext.Annotations,
                statementContext.ContextDescriptor,
                statementContext.ScriptingService,
                false, false, true, false, null);
	        foreach (var node in exprNodes)
	        {
	            // Determine subselects
	            var visitor = new ExprNodeSubselectDeclaredDotVisitor();
	            node.Accept(visitor);

	            // Compile subselects
	            if (!visitor.Subselects.IsEmpty()) {

	                // The outer event type is the filtered-type itself
	                var subselectStreamNumber = 2048;
	                var count = -1;
	                foreach (var subselect in visitor.Subselects) {
	                    count++;
	                    subselectStreamNumber++;
	                    try {
	                        HandleSubselectSelectClauses(subselectStreamNumber, statementContext, subselect,
	                            streamTypeService.EventTypes[0], streamTypeService.StreamNames[0], streamTypeService.StreamNames[0],
	                            taggedEventTypes, arrayEventTypes);
	                    }
	                    catch (ExprValidationException ex) {
	                        throw new ExprValidationException("Failed to validate " + EPStatementStartMethodHelperSubselect.GetSubqueryInfoText(count, subselect) + ": " + ex.Message, ex);
	                    }
	                }
	            }

	            var validated = ExprNodeUtility.GetValidatedSubtree(exprNodeOrigin, node, validationContext);
	            validatedNodes.Add(validated);

	            if ((validated.ExprEvaluator.ReturnType != typeof(bool?)) && ((validated.ExprEvaluator.ReturnType != typeof(bool))))
	            {
	                throw new ExprValidationException("Filter expression not returning a boolean value: '" + ExprNodeUtility.ToExpressionStringMinPrecedence(validated) + "'");
	            }
	        }

	        return validatedNodes;
	    }

	    private static void HandleSubselectSelectClauses(
	        int subselectStreamNumber,
	        StatementContext statementContext,
	        ExprSubselectNode subselect,
	        EventType outerEventType,
	        string outerEventTypeName,
	        string outerStreamName,
	        IDictionary<string, Pair<EventType, string>> taggedEventTypes,
	        IDictionary<string, Pair<EventType, string>> arrayEventTypes)
        {
	        var statementSpec = subselect.StatementSpecCompiled;
	        var filterStreamSpec = statementSpec.StreamSpecs[0];

	        ViewFactoryChain viewFactoryChain;
	        string subselecteventTypeName = null;

	        // construct view factory chain
	        try {
	            if (statementSpec.StreamSpecs[0] is FilterStreamSpecCompiled)
	            {
	                var filterStreamSpecCompiled = (FilterStreamSpecCompiled) statementSpec.StreamSpecs[0];
	                subselecteventTypeName = filterStreamSpecCompiled.FilterSpec.FilterForEventTypeName;

	                // A child view is required to limit the stream
	                if (filterStreamSpec.ViewSpecs.Length == 0)
	                {
	                    throw new ExprValidationException("Subqueries require one or more views to limit the stream, consider declaring a length or time window");
	                }

	                // Register filter, create view factories
	                viewFactoryChain = statementContext.ViewService.CreateFactories(subselectStreamNumber, filterStreamSpecCompiled.FilterSpec.ResultEventType, filterStreamSpec.ViewSpecs, filterStreamSpec.Options, statementContext);
	                subselect.RawEventType = viewFactoryChain.EventType;
	            }
	            else
	            {
	                var namedSpec = (NamedWindowConsumerStreamSpec) statementSpec.StreamSpecs[0];
	                var processor = statementContext.NamedWindowService.GetProcessor(namedSpec.WindowName);
	                viewFactoryChain = statementContext.ViewService.CreateFactories(0, processor.NamedWindowType, namedSpec.ViewSpecs, namedSpec.Options, statementContext);
	                subselecteventTypeName = namedSpec.WindowName;
	            }
	        }
	        catch (ViewProcessingException ex) {
	            throw new ExprValidationException("Error validating subexpression: " + ex.Message, ex);
	        }

	        // the final event type
	        var eventType = viewFactoryChain.EventType;

	        // determine a stream name unless one was supplied
	        var subexpressionStreamName = filterStreamSpec.OptionalStreamName;
	        if (subexpressionStreamName == null)
	        {
	            subexpressionStreamName = "$subselect_" + subselectStreamNumber;
	        }

	        // Named windows don't allow data views
	        if (filterStreamSpec is NamedWindowConsumerStreamSpec)
	        {
	            EPStatementStartMethodHelperValidate.ValidateNoDataWindowOnNamedWindow(viewFactoryChain.FactoryChain);
	        }

	        // Streams event types are the original stream types with the stream zero the subselect stream
	        var namesAndTypes = new LinkedHashMap<string, Pair<EventType, string>>();
	        namesAndTypes.Put(subexpressionStreamName, new Pair<EventType, string>(eventType, subselecteventTypeName));
	        namesAndTypes.Put(outerStreamName, new Pair<EventType, string>(outerEventType, outerEventTypeName));
	        if (taggedEventTypes != null) {
	            foreach (KeyValuePair<string, Pair<EventType, string>> entry in taggedEventTypes) {
	                namesAndTypes.Put(entry.Key, new Pair<EventType, string>(entry.Value.First, entry.Value.Second));
	            }
	        }
	        if (arrayEventTypes != null) {
	            foreach (KeyValuePair<string, Pair<EventType, string>> entry in arrayEventTypes) {
	                namesAndTypes.Put(entry.Key, new Pair<EventType, string>(entry.Value.First, entry.Value.Second));
	            }
	        }
	        StreamTypeService subselectTypeService = new StreamTypeServiceImpl(namesAndTypes, statementContext.EngineURI, true, true);
	        var viewResourceDelegateSubselect = new ViewResourceDelegateUnverified();
	        subselect.FilterSubqueryStreamTypes = subselectTypeService;

	        // Validate select expression
	        var selectClauseSpec = subselect.StatementSpecCompiled.SelectClauseSpec;
	        if (selectClauseSpec.SelectExprList.Length > 0)
	        {
	            if (selectClauseSpec.SelectExprList.Length > 1) {
	                throw new ExprValidationException("Subquery multi-column select is not allowed in this context.");
	            }

	            var element = selectClauseSpec.SelectExprList[0];
	            if (element is SelectClauseExprCompiledSpec)
	            {
	                // validate
	                var compiled = (SelectClauseExprCompiledSpec) element;
	                var selectExpression = compiled.SelectExpression;
	                var evaluatorContextStmt = new ExprEvaluatorContextStatement(statementContext, false);
	                var validationContext = new ExprValidationContext(
	                    subselectTypeService, statementContext.MethodResolutionService, viewResourceDelegateSubselect,
	                    statementContext.SchedulingService, statementContext.VariableService, statementContext.TableService,
	                    evaluatorContextStmt, statementContext.EventAdapterService, statementContext.StatementName,
	                    statementContext.StatementId, statementContext.Annotations, statementContext.ContextDescriptor,
                        statementContext.ScriptingService,
	                    false, false, true, false, null);
	                selectExpression = ExprNodeUtility.GetValidatedSubtree(ExprNodeOrigin.SUBQUERYSELECT, selectExpression, validationContext);
	                subselect.SelectClause = new ExprNode[] {selectExpression};
	                subselect.SelectAsNames = new string[] {compiled.AssignedName};

	                // handle aggregation
	                var aggExprNodes = new List<ExprAggregateNode>();
	                ExprAggregateNodeUtil.GetAggregatesBottomUp(selectExpression, aggExprNodes);
	                if (aggExprNodes.Count > 0)
	                {
	                    // Other stream properties, if there is aggregation, cannot be under aggregation.
	                    foreach (var aggNode in aggExprNodes)
	                    {
	                        var propertiesNodesAggregated = ExprNodeUtility.GetExpressionProperties(aggNode, true);
	                        foreach (var pair in propertiesNodesAggregated)
	                        {
	                            if (pair.First != 0)
	                            {
	                                throw new ExprValidationException("Subselect aggregation function cannot aggregate across correlated properties");
	                            }
	                        }
	                    }

	                    // This stream (stream 0) properties must either all be under aggregation, or all not be.
	                    var propertiesNotAggregated = ExprNodeUtility.GetExpressionProperties(selectExpression, false);
	                    foreach (var pair in propertiesNotAggregated)
	                    {
	                        if (pair.First == 0)
	                        {
	                            throw new ExprValidationException("Subselect properties must all be within aggregation functions");
	                        }
	                    }
	                }
	            }
	        }
	    }

	    public static IList<ExprNode> DecomposeCheckAggregation(IList<ExprNode> validatedNodes)
	    {
	        // Break a top-level AND into constituent expression nodes
	        IList<ExprNode> constituents = new List<ExprNode>();
	        foreach (var validated in validatedNodes)
	        {
	            if (validated is ExprAndNode)
	            {
	                RecursiveAndConstituents(constituents, validated);
	            }
	            else
	            {
	                constituents.Add(validated);
	            }

	            // Ensure there is no aggregation nodes
	            var aggregateExprNodes = new List<ExprAggregateNode>();
	            ExprAggregateNodeUtil.GetAggregatesBottomUp(validated, aggregateExprNodes);
	            if (!aggregateExprNodes.IsEmpty())
	            {
	                throw new ExprValidationException("Aggregation functions not allowed within filters");
	            }
	        }

	        return constituents;
	    }

	    private static void Consolidate(FilterParamExprMap filterParamExprMap, string statementName)
	    {
	        // consolidate or place in a boolean expression (by removing filter spec param from the map)
	        // any filter parameter that feature the same property name and filter operator,
	        // i.e. we are looking for "a!=5 and a!=6"  to transform to "a not in (5,6)" which can match faster
	        // considering that "a not in (5,6) and a not in (7,8)" is "a not in (5, 6, 7, 8)" therefore
	        // we need to consolidate until there is no more work to do
	        IDictionary<Pair<FilterSpecLookupable, FilterOperator>, IList<FilterSpecParam>> mapOfParams = new Dictionary<Pair<FilterSpecLookupable, FilterOperator>, IList<FilterSpecParam>>();

	        bool haveConsolidated;
	        do
	        {
	            haveConsolidated = false;
	            mapOfParams.Clear();

	            // sort into buckets of propertyName + filterOperator combination
	            foreach (var currentParam in filterParamExprMap.FilterParams)
	            {
	                var lookupable = currentParam.Lookupable;
	                var op = currentParam.FilterOperator;
	                var key = new Pair<FilterSpecLookupable, FilterOperator>(lookupable, op);

	                var existingParam = mapOfParams.Get(key);
	                if (existingParam == null)
	                {
	                    existingParam = new List<FilterSpecParam>();
	                    mapOfParams.Put(key, existingParam);
	                }
	                existingParam.Add(currentParam);
	            }

	            foreach (IList<FilterSpecParam> entry in mapOfParams.Values)
	            {
	                if (entry.Count > 1)
	                {
	                    haveConsolidated = true;
	                    Consolidate(entry, filterParamExprMap, statementName);
	                }
	            }
	        }
	        while(haveConsolidated);
	    }

	    // consolidate "val != 3 and val != 4 and val != 5"
	    // to "val not in (3, 4, 5)"
	    private static void HandleConsolidateNotEqual(
	        IList<FilterSpecParam> parameters,
	        FilterParamExprMap filterParamExprMap,
	        string statementName)
	    {
	        IList<FilterSpecParamInValue> values = new List<FilterSpecParamInValue>();

	        ExprNode lastNotEqualsExprNode = null;
	        foreach (var param in parameters)
	        {
	            if (param is FilterSpecParamConstant)
	            {
	                var constantParam = (FilterSpecParamConstant) param;
	                var constant = constantParam.FilterConstant;
	                values.Add(new InSetOfValuesConstant(constant));
	            }
	            else if (param is FilterSpecParamEventProp)
	            {
	                var eventProp = (FilterSpecParamEventProp) param;
	                values.Add(new InSetOfValuesEventProp(eventProp.ResultEventAsName, eventProp.ResultEventProperty,
	                        eventProp.IsMustCoerce, eventProp.CoercionType.GetBoxedType()));
	            }
	            else if (param is FilterSpecParamEventPropIndexed)
	            {
	                var eventProp = (FilterSpecParamEventPropIndexed) param;
	                values.Add(new InSetOfValuesEventPropIndexed(eventProp.ResultEventAsName, eventProp.ResultEventIndex, eventProp.ResultEventProperty,
	                        eventProp.IsMustCoerce, eventProp.CoercionType.GetBoxedType(), statementName));
	            }
	            else
	            {
	                throw new ArgumentException("Unknown filter parameter:" + param.ToString());
	            }

	            lastNotEqualsExprNode = filterParamExprMap.RemoveEntry(param);
	        }

	        var paramX = new FilterSpecParamIn(parameters[0].Lookupable, FilterOperator.NOT_IN_LIST_OF_VALUES, values);
	        filterParamExprMap.Put(lastNotEqualsExprNode, paramX);
	    }

        /// <summary>
        /// For a given expression determine if this is optimizable and create the filter parameter
        /// representing the expression, or null if not optimizable.
        /// </summary>
        /// <param name="constituent">is the expression to look at</param>
        /// <param name="taggedEventTypes">event types and their tags</param>
        /// <param name="arrayEventTypes">@return filter parameter representing the expression, or null</param>
        /// <param name="exprEvaluatorContext">The expr evaluator context.</param>
        /// <param name="statementName">Name of the statement.</param>
        /// <returns>
        /// FilterSpecParam filter param
        /// </returns>
        /// <throws>ExprValidationException if the expression is invalid</throws>
	    public static FilterSpecParam MakeFilterParam(
	        ExprNode constituent,
            IDictionary<string, Pair<EventType, string>> taggedEventTypes,
            IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        ExprEvaluatorContext exprEvaluatorContext,
	        string statementName)
	    {
	        // Is this expresson node a simple compare, i.e. a=5 or b<4; these can be indexed
	        if ((constituent is ExprEqualsNode) ||
	            (constituent is ExprRelationalOpNode))
	        {
	            var param = HandleEqualsAndRelOp(constituent, arrayEventTypes, exprEvaluatorContext, statementName);
	            if (param != null)
	            {
	                return param;
	            }
	        }

	        constituent = RewriteOrToInIfApplicable(constituent);

	        // Is this expression node a simple compare, i.e. a=5 or b<4; these can be indexed
	        if (constituent is ExprInNode)
	        {
	            var param = HandleInSetNode((ExprInNode)constituent, arrayEventTypes, exprEvaluatorContext, statementName);
	            if (param != null)
	            {
	                return param;
	            }
	        }

	        if (constituent is ExprBetweenNode)
	        {
	            var param = HandleRangeNode((ExprBetweenNode)constituent, arrayEventTypes, exprEvaluatorContext, statementName);
	            if (param != null)
	            {
	                return param;
	            }
	        }

	        if (constituent is ExprPlugInSingleRowNode) {
	            var param = HandlePlugInSingleRow((ExprPlugInSingleRowNode) constituent, exprEvaluatorContext, statementName);
	            if (param != null)
	            {
	                return param;
	            }
	        }

	        return null;
	    }

	    public static ExprNode RewriteOrToInIfApplicable(ExprNode constituent)
        {
	        if (!(constituent is ExprOrNode) || constituent.ChildNodes.Length < 2) {
	            return constituent;
	        }

	        // check eligibility
	        var childNodes = constituent.ChildNodes;
	        foreach (var child in childNodes) {
	            if (!(child is ExprEqualsNode)) {
	                return constituent;
	            }
	            var equalsNode = (ExprEqualsNode) child;
	            if (equalsNode.IsIs || equalsNode.IsNotEquals) {
	                return constituent;
	            }
	        }

	        // find common-expression node
	        ExprNode commonExpressionNode;
	        var lhs = childNodes[0].ChildNodes[0];
	        var rhs = childNodes[0].ChildNodes[1];
	        if (ExprNodeUtility.DeepEquals(lhs, rhs)) {
	            return constituent;
	        }
	        if (IsExprExistsInAllEqualsChildNodes(childNodes, lhs)) {
	            commonExpressionNode = lhs;
	        }
	        else if (IsExprExistsInAllEqualsChildNodes(childNodes, rhs)) {
	            commonExpressionNode = rhs;
	        }
	        else {
	            return constituent;
	        }

	        // build node
	        var @in = new ExprInNodeImpl(false);
	        @in.AddChildNode(commonExpressionNode);
	        for (var i = 0; i < constituent.ChildNodes.Length; i++) {
	            var child = constituent.ChildNodes[i];
	            var nodeindex = ExprNodeUtility.DeepEquals(commonExpressionNode, childNodes[i].ChildNodes[0]) ? 1 : 0;
	            @in.AddChildNode(child.ChildNodes[nodeindex]);
	        }

	        // validate
	        try {
	            @in.ValidateWithoutContext();
	        }
	        catch (ExprValidationException) {
	            return constituent;
	        }

	        return @in;
	    }

	    private static FilterSpecParam HandlePlugInSingleRow(
	        ExprPlugInSingleRowNode constituent,
	        ExprEvaluatorContext exprEvaluatorContext,
	        string statementName)
        {
	        if (constituent.ExprEvaluator.ReturnType.GetBoxedType() != typeof(bool?)) {
	            return null;
	        }
	        if (!constituent.IsFilterLookupEligible) {
	            return null;
	        }
	        var lookupable = constituent.FilterLookupable;
	        return new FilterSpecParamConstant(lookupable, FilterOperator.EQUAL, true);
	    }

	    private static FilterSpecParam HandleRangeNode(
	        ExprBetweenNode betweenNode,
            IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        ExprEvaluatorContext exprEvaluatorContext,
	        string statementName)
	    {
	        var left = betweenNode.ChildNodes[0];
	        if (left is ExprFilterOptimizableNode)
	        {
	            var filterOptimizableNode = (ExprFilterOptimizableNode) left;
	            var lookupable = filterOptimizableNode.FilterLookupable;
	            FilterOperator op = FilterOperatorExtensions.ParseRangeOperator(betweenNode.IsLowEndpointIncluded, betweenNode.IsHighEndpointIncluded,
	                    betweenNode.IsNotBetween);

	            var low = HandleRangeNodeEndpoint(betweenNode.ChildNodes[1], arrayEventTypes, exprEvaluatorContext, statementName);
	            var high = HandleRangeNodeEndpoint(betweenNode.ChildNodes[2], arrayEventTypes, exprEvaluatorContext, statementName);

	            if ((low != null) && (high != null))
	            {
	                return new FilterSpecParamRange(lookupable, op, low, high);
	            }
	        }
	        return null;
	    }

	    private static FilterSpecParamRangeValue HandleRangeNodeEndpoint(
	        ExprNode endpoint,
	        IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        ExprEvaluatorContext exprEvaluatorContext,
	        string statementName)
	    {
	        // constant
	        if (ExprNodeUtility.IsConstantValueExpr(endpoint))
	        {
	            var node = (ExprConstantNode) endpoint;
	            var value = node.GetConstantValue(exprEvaluatorContext);
	            if (value is string) {
	                return new RangeValueString((string)value);
	            }
	            else {
	                return new RangeValueDouble(value.AsDouble());
	            }
	        }

	        if (endpoint is ExprContextPropertyNode) {
	            var node = (ExprContextPropertyNode) endpoint;
	            return new RangeValueContextProp(node.Getter);
	        }

	        // or property
	        if (endpoint is ExprIdentNode)
	        {
	            var identNodeInner = (ExprIdentNode) endpoint;
	            if (identNodeInner.StreamId == 0)
	            {
	                return null;
	            }

	            if (arrayEventTypes != null && !arrayEventTypes.IsEmpty() && arrayEventTypes.ContainsKey(identNodeInner.ResolvedStreamName))
	            {
	                var indexAndProp = GetStreamIndex(identNodeInner.ResolvedPropertyName);
	                return new RangeValueEventPropIndexed(identNodeInner.ResolvedStreamName, indexAndProp.First, indexAndProp.Second, statementName);
	            }
	            else
	            {
	                return new RangeValueEventProp(identNodeInner.ResolvedStreamName, identNodeInner.ResolvedPropertyName);
	            }
	        }

	        return null;
	    }

	    private static FilterSpecParam HandleInSetNode(
	        ExprInNode constituent,
	        IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        ExprEvaluatorContext exprEvaluatorContext,
	        string statementName)
	    {
	        var left = constituent.ChildNodes[0];
	        if (!(left is ExprFilterOptimizableNode)) {
	            return null;
	        }

	        var filterOptimizableNode = (ExprFilterOptimizableNode) left;
	        var lookupable = filterOptimizableNode.FilterLookupable;
	        var op = FilterOperator.IN_LIST_OF_VALUES;
	        if (constituent.IsNotIn)
	        {
	            op = FilterOperator.NOT_IN_LIST_OF_VALUES;
	        }

	        var expectedNumberOfConstants = constituent.ChildNodes.Length - 1;
            var listofValues = new List<FilterSpecParamInValue>();
            var it = ((IList<ExprNode>)constituent.ChildNodes).GetEnumerator();
            var evaluateParams = new EvaluateParams(null, true, exprEvaluatorContext);
	        
            it.MoveNext();  // ignore the first node as it's the identifier
	        while (it.MoveNext())
	        {
	            var subNode = it.Current;
	            if (ExprNodeUtility.IsConstantValueExpr(subNode))
	            {
	                var constantNode = (ExprConstantNode) subNode;
	                var constant = constantNode.GetConstantValue(exprEvaluatorContext);
	                if (constant == null)
	                {
	                    return null;
	                }
	                else if (constant.GetType().IsArray)
	                {
	                    var asArray = constant as Array;
	                    for (var i = 0; i < asArray.Length; i++)
	                    {
	                        var arrayElement = asArray.GetValue(i);
	                        var arrayElementCoerced = HandleConstantsCoercion(lookupable, arrayElement);
	                        listofValues.Add(new InSetOfValuesConstant(arrayElementCoerced));
	                        if (i > 0)
	                        {
	                            expectedNumberOfConstants++;
	                        }
	                    }
	                }
	                else if (constant.GetType().IsGenericCollection())
	                {
	                    return null;
	                }
	                else if (constant.GetType().IsGenericDictionary())
	                {
	                    return null;
	                }
	                else
	                {
	                    constant = HandleConstantsCoercion(lookupable, constant);
	                    listofValues.Add(new InSetOfValuesConstant(constant));
	                }
	            }
	            if (subNode is ExprContextPropertyNode)
	            {
	                var contextPropertyNode = (ExprContextPropertyNode) subNode;
	                var returnType = contextPropertyNode.ReturnType;
	                if (contextPropertyNode.ReturnType.IsImplementsInterface(typeof (ICollection<>)) ||
	                    contextPropertyNode.ReturnType.IsImplementsInterface(typeof (IDictionary<,>)))
	                {
	                    return null;
	                }
	                if ((returnType != null) && (returnType.GetType().IsArray))
	                {
	                    return null;
	                }
	                var coercer = GetNumberCoercer(
	                    left.ExprEvaluator.ReturnType, contextPropertyNode.ReturnType, lookupable.Expression);
	                listofValues.Add(
	                    new InSetOfValuesContextProp(contextPropertyNode.PropertyName, contextPropertyNode.Getter, coercer));
	            }
	            if (subNode is ExprIdentNode)
	            {
	                var identNodeInner = (ExprIdentNode) subNode;
	                if (identNodeInner.StreamId == 0)
	                {
	                    break; // for same event evals use the boolean expression, via count compare failing below
	                }

	                var isMustCoerce = false;
	                var numericCoercionType = lookupable.ReturnType.GetBoxedType();
	                if (identNodeInner.ExprEvaluator.ReturnType != lookupable.ReturnType)
	                {
	                    if (lookupable.ReturnType.IsNumeric())
	                    {
	                        if (!identNodeInner.ExprEvaluator.ReturnType.CanCoerce(lookupable.ReturnType))
	                        {
	                            ThrowConversionError(identNodeInner.ExprEvaluator.ReturnType, lookupable.ReturnType, lookupable.Expression);
	                        }
	                        isMustCoerce = true;
	                    }
	                    else {
	                        break;  // assumed not compatible
	                    }
	                }

	                FilterSpecParamInValue inValue;
	                var streamName = identNodeInner.ResolvedStreamName;
	                if (arrayEventTypes != null && !arrayEventTypes.IsEmpty() && arrayEventTypes.ContainsKey(streamName))
	                {
	                    var indexAndProp = GetStreamIndex(identNodeInner.ResolvedPropertyName);
	                    inValue = new InSetOfValuesEventPropIndexed(identNodeInner.ResolvedStreamName, indexAndProp.First,
	                            indexAndProp.Second, isMustCoerce, numericCoercionType, statementName);
	                }
	                else
	                {
	                    inValue = new InSetOfValuesEventProp(identNodeInner.ResolvedStreamName, identNodeInner.ResolvedPropertyName, isMustCoerce, numericCoercionType);
	                }

	                listofValues.Add(inValue);
	            }
	        }

	        // Fallback if not all values in the in-node can be resolved to properties or constants
	        if (listofValues.Count == expectedNumberOfConstants)
	        {
	            return new FilterSpecParamIn(lookupable, op, listofValues);
	        }
	        return null;
	    }

	    private static FilterSpecParam HandleEqualsAndRelOp(
	        ExprNode constituent,
	        IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        ExprEvaluatorContext exprEvaluatorContext,
	        string statementName)
	    {
	        FilterOperator op;
	        if (constituent is ExprEqualsNode)
	        {
	            var equalsNode = (ExprEqualsNode) constituent;
	            if (!equalsNode.IsIs) {
	                op = FilterOperator.EQUAL;
	                if (equalsNode.IsNotEquals)
	                {
	                    op = FilterOperator.NOT_EQUAL;
	                }
	            }
	            else {
	                op = FilterOperator.IS;
	                if (equalsNode.IsNotEquals)
	                {
	                    op = FilterOperator.IS_NOT;
	                }
	            }
	        }
	        else
	        {
	            var relNode = (ExprRelationalOpNode) constituent;
	            if (relNode.RelationalOpEnum == RelationalOpEnum.GT)
	            {
	                op = FilterOperator.GREATER;
	            }
	            else if (relNode.RelationalOpEnum == RelationalOpEnum.LT)
	            {
	                op = FilterOperator.LESS;
	            }
	            else if (relNode.RelationalOpEnum == RelationalOpEnum.LE)
	            {
	                op = FilterOperator.LESS_OR_EQUAL;
	            }
	            else if (relNode.RelationalOpEnum == RelationalOpEnum.GE)
	            {
	                op = FilterOperator.GREATER_OR_EQUAL;
	            }
	            else
	            {
	                throw new IllegalStateException("Opertor '" + relNode.RelationalOpEnum + "' not mapped");
	            }
	        }

	        var left = constituent.ChildNodes[0];
	        var right = constituent.ChildNodes[1];

	        // check identifier and constant combination
	        if ((ExprNodeUtility.IsConstantValueExpr(right)) && (left is ExprFilterOptimizableNode))
	        {
	            var filterOptimizableNode = (ExprFilterOptimizableNode) left;
	            if (filterOptimizableNode.IsFilterLookupEligible) {
	                var constantNode = (ExprConstantNode) right;
	                var lookupable = filterOptimizableNode.FilterLookupable;
	                var constant = constantNode.GetConstantValue(exprEvaluatorContext);
	                constant = HandleConstantsCoercion(lookupable, constant);
	                return new FilterSpecParamConstant(lookupable, op, constant);
	            }
	        }
	        if ((ExprNodeUtility.IsConstantValueExpr(left)) && (right is ExprFilterOptimizableNode))
	        {
	            var filterOptimizableNode = (ExprFilterOptimizableNode) right;
	            if (filterOptimizableNode.IsFilterLookupEligible) {
	                var constantNode = (ExprConstantNode) left;
	                var lookupable = filterOptimizableNode.FilterLookupable;
	                var constant = constantNode.GetConstantValue(exprEvaluatorContext);
	                constant = HandleConstantsCoercion(lookupable, constant);
	                var opReversed = op.IsComparisonOperator() ? op.ReversedRelationalOp() : op;
	                return new FilterSpecParamConstant(lookupable, opReversed, constant);
	            }
	        }
	        // check identifier and expression containing other streams
	        if ((left is ExprIdentNode) && (right is ExprIdentNode))
	        {
	            var identNodeLeft = (ExprIdentNode) left;
	            var identNodeRight = (ExprIdentNode) right;

	            if ((identNodeLeft.StreamId == 0) && (identNodeLeft.IsFilterLookupEligible) && (identNodeRight.StreamId != 0))
	            {
	                return HandleProperty(op, identNodeLeft, identNodeRight, arrayEventTypes, statementName);
	            }
	            if ((identNodeRight.StreamId == 0) && (identNodeRight.IsFilterLookupEligible) && (identNodeLeft.StreamId != 0))
	            {
	                op = GetReversedOperator(constituent, op); // reverse operators, as the expression is "stream1.prop xyz stream0.prop"
	                return HandleProperty(op, identNodeRight, identNodeLeft, arrayEventTypes, statementName);
	            }
	        }

	        if ((left is ExprFilterOptimizableNode) && (right is ExprContextPropertyNode)) {
	            var filterOptimizableNode = (ExprFilterOptimizableNode) left;
	            var ctxNode = (ExprContextPropertyNode) right;
	            var lookupable = filterOptimizableNode.FilterLookupable;
	            if (filterOptimizableNode.IsFilterLookupEligible) {
	                var numberCoercer = GetNumberCoercer(lookupable.ReturnType, ctxNode.ReturnType, lookupable.Expression);
	                return new FilterSpecParamContextProp(lookupable, op, ctxNode.PropertyName, ctxNode.Getter, numberCoercer);
	            }
	        }
	        if ((left is ExprContextPropertyNode) && (right is ExprFilterOptimizableNode)) {
	            var filterOptimizableNode = (ExprFilterOptimizableNode) right;
	            var ctxNode = (ExprContextPropertyNode) left;
	            var lookupable = filterOptimizableNode.FilterLookupable;
	            if (filterOptimizableNode.IsFilterLookupEligible) {
	                op = GetReversedOperator(constituent, op); // reverse operators, as the expression is "stream1.prop xyz stream0.prop"
	                var numberCoercer = GetNumberCoercer(lookupable.ReturnType, ctxNode.ReturnType, lookupable.Expression);
	                return new FilterSpecParamContextProp(lookupable, op, ctxNode.PropertyName, ctxNode.Getter, numberCoercer);
	            }
	        }
	        return null;
	    }

	    private static FilterOperator GetReversedOperator(ExprNode constituent, FilterOperator op) {
	        if (!(constituent is ExprRelationalOpNode)) {
	            return op;
	        }

	        var relNode = (ExprRelationalOpNode) constituent;
	        var relationalOpEnum = relNode.RelationalOpEnum;

	        if (relationalOpEnum == RelationalOpEnum.GT)
	        {
	            return FilterOperator.LESS;
	        }
	        else if (relationalOpEnum == RelationalOpEnum.LT)
	        {
	            return FilterOperator.GREATER;
	        }
	        else if (relationalOpEnum == RelationalOpEnum.LE)
	        {
	            return FilterOperator.GREATER_OR_EQUAL;
	        }
	        else if (relationalOpEnum == RelationalOpEnum.GE)
	        {
	            return FilterOperator.LESS_OR_EQUAL;
	        }
	        return op;
	    }

	    private static FilterSpecParam HandleProperty(
	        FilterOperator op,
	        ExprIdentNode identNodeLeft,
	        ExprIdentNode identNodeRight,
            IDictionary<string, Pair<EventType, string>> arrayEventTypes,
	        string statementName)
	    {
	        var propertyName = identNodeLeft.ResolvedPropertyName;

	        Type leftType = identNodeLeft.ExprEvaluator.ReturnType;
	        Type rightType = identNodeRight.ExprEvaluator.ReturnType;

	        var numberCoercer = GetNumberCoercer(leftType, rightType, propertyName);
	        var isMustCoerce = numberCoercer != null;
	        var numericCoercionType = leftType.GetBoxedType();

	        var streamName = identNodeRight.ResolvedStreamName;
	        if (arrayEventTypes != null && !arrayEventTypes.IsEmpty() && arrayEventTypes.ContainsKey(streamName))
	        {
	            var indexAndProp = GetStreamIndex(identNodeRight.ResolvedPropertyName);
	            return new FilterSpecParamEventPropIndexed(identNodeLeft.FilterLookupable, op, identNodeRight.ResolvedStreamName, indexAndProp.First,
	                    indexAndProp.Second, isMustCoerce, numberCoercer, numericCoercionType, statementName);
	        }
	        return new FilterSpecParamEventProp(identNodeLeft.FilterLookupable, op, identNodeRight.ResolvedStreamName, identNodeRight.ResolvedPropertyName,
	                isMustCoerce, numberCoercer, numericCoercionType, statementName);
	    }

	    private static Coercer GetNumberCoercer(Type leftType, Type rightType, string expression)
        {
	        var numericCoercionType = leftType.GetBoxedType();
	        if (rightType != leftType)
	        {
	            if (rightType.IsNumeric())
	            {
	                if (!rightType.CanCoerce(leftType))
	                {
	                    ThrowConversionError(rightType, leftType, expression);
	                }
	                return CoercerFactory.GetCoercer(rightType, numericCoercionType);
	            }
	        }
	        return null;
	    }

	    private static Pair<int, string> GetStreamIndex(string resolvedPropertyName)
	    {
	        var property = PropertyParser.ParseAndWalk(resolvedPropertyName);
	        if (!(property is NestedProperty))
	        {
	            throw new IllegalStateException("Expected a nested property providing an index for array match '" + resolvedPropertyName + "'");
	        }
	        var nested = ((NestedProperty) property);
	        if (nested.Properties.Count < 2)
	        {
	            throw new IllegalStateException("Expected a nested property name for array match '" + resolvedPropertyName + "', none found");
	        }
	        if (!(nested.Properties[0] is IndexedProperty))
	        {
	            throw new IllegalStateException("Expected an indexed property for array match '" + resolvedPropertyName + "', please provide an index");
	        }
	        var index = ((IndexedProperty) nested.Properties[0]).Index;
	        nested.Properties.RemoveAt(0);
	        var writer = new StringWriter();
	        nested.ToPropertyEPL(writer);
	        return new Pair<int, string>(index, writer.ToString());
	    }

	    private static void ThrowConversionError(Type fromType, Type toType, string propertyName)
	    {
	        var text = string.Format("Implicit conversion from datatype '{0}' to '{1}' for property '{2}' is not allowed (strict filter type coercion)", fromType.FullName, toType.FullName, propertyName);
	        throw new ExprValidationException(text);
	    }

	    // expressions automatically coerce to the most upwards type
	    // filters require the same type
	    private static object HandleConstantsCoercion(FilterSpecLookupable lookupable, object constant)
	    {
	        var identNodeType = lookupable.ReturnType;
	        if (!identNodeType.IsNumeric())
	        {
	            return constant;    // no coercion required, other type checking performed by expression this comes from
	        }

	        if (constant == null)  // null constant type
	        {
	            return null;
	        }

	        if (!constant.GetType().CanCoerce(identNodeType))
	        {
	            ThrowConversionError(constant.GetType(), identNodeType, lookupable.Expression);
	        }

	        var identNodeTypeBoxed = identNodeType.GetBoxedType();
	        return CoercerFactory.CoerceBoxed(constant, identNodeTypeBoxed);
	    }

	    private static void RecursiveAndConstituents(IList<ExprNode> constituents, ExprNode exprNode)
	    {
	        foreach (var inner in exprNode.ChildNodes)
	        {
	            if (inner is ExprAndNode)
	            {
	                RecursiveAndConstituents(constituents, inner);
	            }
	            else
	            {
	                constituents.Add(inner);
	            }
	        }
	    }

	    private static bool IsExprExistsInAllEqualsChildNodes(ExprNode[] childNodes, ExprNode search)
	    {
	        foreach (var child in childNodes)
	        {
	            var lhs = child.ChildNodes[0];
	            var rhs = child.ChildNodes[1];
	            if (!ExprNodeUtility.DeepEquals(lhs, search) && !ExprNodeUtility.DeepEquals(rhs, search))
	            {
	                return false;
	            }
	            if (ExprNodeUtility.DeepEquals(lhs, rhs))
	            {
	                return false;
	            }
	        }
	        return true;
	    }
	}
} // end of namespace
